import numpy as np
import xarray as xr

from xrspatial import a_star_search


def test_a_star_search():
    agg = xr.DataArray(np.array([[0, 1, 0, 0],
                                 [1, 1, 0, 0],
                                 [0, 1, 2, 2],
                                 [1, 0, 2, 0],
                                 [0, 2, 2, 2]]),
                       dims=['lat', 'lon'])

    height, width = agg.shape
    _lon = np.linspace(0, width - 1, width)
    _lat = np.linspace(0, height - 1, height)
    agg['lon'] = _lon
    agg['lat'] = _lat
    barriers = []
    # no barriers, there always path from a start location to a goal location
    for x0 in _lon:
        for y0 in _lat:
            start = (x0, y0)
            for x1 in _lon:
                for y1 in _lat:
                    goal = (x1, y1)
                    path_agg = a_star_search(agg, start, goal, barriers,
                                             'lon', 'lat')
                    assert isinstance(path_agg, xr.DataArray)
                    assert type(path_agg.values[0][0]) == np.float64
                    assert path_agg.shape == agg.shape
                    assert path_agg.dims == agg.dims
                    assert path_agg.attrs == agg.attrs
                    for c in path_agg.coords:
                        assert (path_agg[c] == agg.coords[c]).all()
                    if start == goal:
                        assert np.nanmax(path_agg) == 0 and \
                               np.nanmin(path_agg) == 0
                    else:
                        assert np.nanmax(path_agg) > 0 and \
                               np.nanmin(path_agg) == 0

    barriers = [1]
    # set pixels with value 1 as barriers,
    # cannot go from (0, 0) to anywhere since it is surrounded by 1s
    start = (0, 0)
    for x1 in _lon:
        for y1 in _lat:
            goal = (x1, y1)
            if goal != start:
                path_agg = a_star_search(agg, start, goal, barriers,
                                         'lon', 'lat')
                assert isinstance(path_agg, xr.DataArray)
                assert type(path_agg.values[0][0]) == np.float64
                assert path_agg.shape == agg.shape
                assert path_agg.dims == agg.dims
                assert path_agg.attrs == agg.attrs
                for c in path_agg.coords:
                    assert (path_agg[c] == agg.coords[c]).all()
                # no path, all cells in path_agg are nans
                assert np.isnan(path_agg).all()

    # test with nans
    agg = xr.DataArray(np.array([[0, 1, 0, 0],
                                 [1, 1, np.nan, 0],
                                 [0, 1, 2, 2],
                                 [1, 0, 2, 0],
                                 [0, np.nan, 2, 2]]),
                       dims=['lat', 'lon'])

    height, width = agg.shape
    _lon = np.linspace(0, width - 1, width)
    _lat = np.linspace(0, height - 1, height)
    agg['lon'] = _lon
    agg['lat'] = _lat
    # start and end at a nan pixel, coordinate in (lon, lat) format
    # in this example, each pixel is a unit of lon and lat,
    #                  start = (2, 1) corresponds to pixel at (1, 2),
    #                  goal = (1, 4) corresponds to pixel at (4, 1)
    start = (2, 1)
    goal = (1, 4)
    # no barriers
    barriers = []
    # no snap
    no_snap_path_agg = a_star_search(agg, start, goal, barriers, 'lon', 'lat')
    # no path, all cells in path_agg are nans
    assert np.isnan(no_snap_path_agg).all()

    # set snap_start = True, snap_goal = False
    snap_start_path_agg = a_star_search(agg, start, goal, barriers,
                                        'lon', 'lat', snap_start=True)
    # no path, all cells in path_agg are nans
    assert np.isnan(snap_start_path_agg).all()

    # set snap_start = False, snap_goal = True
    snap_goal_path_agg = a_star_search(agg, start, goal, barriers,
                                       'lon', 'lat', snap_goal=True)
    # no path, all cells in path_agg are nans
    assert np.isnan(snap_goal_path_agg).all()

    # set snap_start = True, snap_goal = True
    # 8-connectivity as default
    path_agg_8 = a_star_search(agg, start, goal, barriers, 'lon', 'lat',
                               snap_start=True, snap_goal=True)
    # path exists
    expected_result_8 = np.array([[np.nan, np.nan, 0., np.nan],
                                 [np.nan, 1.41421356, np.nan, np.nan],
                                 [np.nan, 2.41421356, np.nan, np.nan],
                                 [np.nan, 3.41421356, np.nan, np.nan],
                                 [np.nan, np.nan, np.nan, np.nan]])
    assert ((np.isnan(path_agg_8) & np.isnan(expected_result_8)) | (
                    abs(path_agg_8 - expected_result_8) <= 1e-5)).all()

    # 4-connectivity
    path_agg_4 = a_star_search(agg, start, goal, barriers, 'lon', 'lat',
                               snap_start=True, snap_goal=True, connectivity=4)
    # path exists
    expected_result_4 = np.array([[np.nan, 1, 0., np.nan],
                                  [np.nan, 2, np.nan, np.nan],
                                  [np.nan, 3, np.nan, np.nan],
                                  [np.nan, 4, np.nan, np.nan],
                                  [np.nan, np.nan, np.nan, np.nan]])
    assert ((np.isnan(path_agg_4) & np.isnan(expected_result_4)) | (
                    abs(path_agg_4 - expected_result_4) <= 1e-5)).all()



import dask.array as da
import xarray as xr

from xrspatial.datasets import make_terrain


def test_make_terrain():
    terrain = make_terrain()
    assert terrain is not None
    assert isinstance(terrain, xr.DataArray)
    assert isinstance(terrain.data, da.Array)



from math import ceil

import dask.array as da
import datashader as ds
import datashader.transfer_functions as tf
from numba import cuda, jit
import numpy as np
import xarray as xr


try:
    import cupy

    if cupy.result_type is np.result_type:
        # hack until cupy release of https://github.com/cupy/cupy/pull/2249
        # Without this, cupy.histogram raises an error that cupy.result_type
        # is not defined.
        cupy.result_type = lambda *args: np.result_type(
            *[
                arg.dtype if isinstance(arg, cupy.ndarray)
                else arg for arg in args
            ]
        )
except ImportError:
    cupy = None

ngjit = jit(nopython=True, nogil=True)


def has_cuda():
    """Check for supported CUDA device. If none found, return False"""
    local_cuda = False
    try:
        cuda.cudadrv.devices.gpus.current
        local_cuda = True
    except cuda.cudadrv.error.CudaSupportError:
        local_cuda = False

    return local_cuda


def doesnt_have_cuda():
    return not has_cuda()


def cuda_args(shape):
    """
    Compute the blocks-per-grid and threads-per-block parameters for
    use when invoking cuda kernels

    Parameters
    ----------
    shape: int or tuple of ints
        The shape of the input array that the kernel will parallelize
        over.

    Returns
    -------
    bpg, tpb : tuple
        Tuple of (blocks_per_grid, threads_per_block).
    """
    if isinstance(shape, int):
        shape = (shape,)

    max_threads = cuda.get_current_device().MAX_THREADS_PER_BLOCK
    # Note: We divide max_threads by 2.0 to leave room for the registers
    threads_per_block = int(ceil(max_threads / 2.0) ** (1.0 / len(shape)))
    tpb = (threads_per_block,) * len(shape)
    bpg = tuple(int(ceil(d / threads_per_block)) for d in shape)
    return bpg, tpb


def is_cupy_backed(agg: xr.DataArray):
    try:
        return type(agg.data._meta).__module__.split(".")[0] == "cupy"
    except AttributeError:
        return False


def is_dask_cupy(agg: xr.DataArray):
    return isinstance(agg.data, da.Array) and is_cupy_backed(agg)


class ArrayTypeFunctionMapping(object):
    def __init__(self, numpy_func, cupy_func, dask_func, dask_cupy_func):
        self.numpy_func = numpy_func
        self.cupy_func = cupy_func
        self.dask_func = dask_func
        self.dask_cupy_func = dask_cupy_func

    def __call__(self, arr):

        # numpy case
        if isinstance(arr.data, np.ndarray):
            return self.numpy_func

        # cupy case
        elif has_cuda() and isinstance(arr.data, cupy.ndarray):
            return self.cupy_func

        # dask + cupy case
        elif has_cuda() and is_dask_cupy(arr):
            return self.dask_cupy_func

        # dask + numpy case
        elif isinstance(arr.data, da.Array):
            return self.dask_func

        else:
            raise TypeError("Unsupported Array Type: {}".format(type(arr)))


def validate_arrays(*arrays):
    if len(arrays) < 2:
        raise ValueError(
            "validate_arrays() input must contain 2 or more arrays"
        )

    first_array = arrays[0]
    for i in range(1, len(arrays)):

        if not first_array.data.shape == arrays[i].data.shape:
            raise ValueError("input arrays must have equal shapes")

        if not isinstance(first_array.data, type(arrays[i].data)):
            raise ValueError("input arrays must have same type")


def calc_res(raster):
    """
    Calculate the resolution of xarray.DataArray raster and return it
    as thetwo-tuple (xres, yres).

    Parameters
    ----------
    raster: xr.DataArray
        Input raster.

    Returns
    -------
    xres, yres: tuple
        Tuple of (x-resolution, y-resolution).

    Notes
    -----
        - Sourced from datashader.utils.
    """
    h, w = raster.shape[-2:]
    ydim, xdim = raster.dims[-2:]
    xcoords = raster[xdim].values
    ycoords = raster[ydim].values
    xres = (xcoords[-1] - xcoords[0]) / (w - 1)
    yres = (ycoords[0] - ycoords[-1]) / (h - 1)
    return xres, yres


def get_dataarray_resolution(agg: xr.DataArray):
    """
    Calculate resolution of xarray.DataArray.

    Parameters
    ----------
    agg: xarray.DataArray
        Input raster.

    Returns
    -------
    cellsize_x, cellsize_y: tuple
        Tuple of (x cell size, y cell size).
    """
    # get cellsize out from 'res' attribute
    try:
        cellsize = agg.attrs.get("res")
        if (
            isinstance(cellsize, (tuple, np.ndarray, list))
            and len(cellsize) == 2
            and isinstance(cellsize[0], (int, float))
            and isinstance(cellsize[1], (int, float))
        ):
            cellsize_x, cellsize_y = cellsize
        elif isinstance(cellsize, (int, float)):
            cellsize_x = cellsize
            cellsize_y = cellsize
        else:
            cellsize_x, cellsize_y = calc_res(agg)

    except Exception:
        cellsize_x, cellsize_y = calc_res(agg)

    return cellsize_x, cellsize_y


def lnglat_to_meters(longitude, latitude):
    """
    Projects the given (longitude, latitude) values into Web Mercator
    coordinates (meters East of Greenwich and meters North of the
    Equator).

    Longitude and latitude can be provided as scalars, Pandas columns,
    or Numpy arrays, and will be returned in the same form.  Lists
    or tuples will be converted to Numpy arrays.

    Parameters
    ----------
    latitude: float
        Input latitude.
    longitude: float
        Input longitude.

    Returns
    -------
    easting, northing : tuple
        Tuple of (easting, northing).

    Examples
    --------
    .. sourcecode:: python

        >>> easting, northing = lnglat_to_meters(-40.71,74)
        >>> easting, northing = lnglat_to_meters(np.array([-74]),
        >>>                                      np.array([40.71]))
        >>> df = pandas.DataFrame(dict(longitude=np.array([-74]),
        >>>                            latitude=np.array([40.71])))
        >>> df.loc[:, 'longitude'], df.loc[:, 'latitude'] = lnglat_to_meters(
        >>>     df.longitude, df.latitude)
    """
    if isinstance(longitude, (list, tuple)):
        longitude = np.array(longitude)
    if isinstance(latitude, (list, tuple)):
        latitude = np.array(latitude)

    origin_shift = np.pi * 6378137
    easting = longitude * origin_shift / 180.0
    northing = np.log(
        np.tan((90 + latitude) * np.pi / 360.0)
        ) * origin_shift / np.pi
    return (easting, northing)


def height_implied_by_aspect_ratio(W, X, Y):
    """
    Utility function for calculating height (in pixels) which is implied
    by a width, x-range, and y-range. Simple ratios are used to maintain
    aspect ratio.

    Parameters
    ----------
    W: int
        Width in pixel.
    X: tuple
        X-range in data units.
    Y: tuple
        X-range in data units.

    Returns
    -------
    height : int
        height in pixels

    Examples
    --------
    .. sourcecode:: python

        >>> plot_width = 1000
        >>> x_range = (0,35
        >>> y_range = (0, 70)
        >>> plot_height = height_implied_by_aspect_ratio(
                plot_width,
                x_range,
                y_range,
            )
    """
    return int((W * (Y[1] - Y[0])) / (X[1] - X[0]))


def bands_to_img(r, g, b, nodata=1):
    h, w = r.shape
    data = np.zeros((h, w, 4), dtype=np.uint8)
    data[:, :, 0] = (r).astype(np.uint8)
    data[:, :, 1] = (g).astype(np.uint8)
    data[:, :, 2] = (b).astype(np.uint8)
    a = np.where(np.logical_or(np.isnan(r), r <= nodata), 0, 255)
    data[:, :, 3] = a.astype(np.uint8)
    return tf.Image.fromarray(data, "RGBA")


def canvas_like(
    raster,
    width=512,
    height=None,
    x_range=None,
    y_range=None,
    **kwargs
):

    """
    Resample a xarray.DataArray by canvas width and bounds.
    Height of the resampled raster is implied from the canvas width
    using aspect ratio of original raster.

    This function uses of datashader.Canvas.raster internally.
    Most of the docstrings are copied from Datashader.

    Handles 2D or 3D xarray.DataArray, assuming that the last two
    array dimensions are the y-axis and x-axis that are to be
    resampled. If a 3D array is supplied a layer may be specified
    to resample to select the layer along the first dimension to
    resample.

    Parameters
    ----------
    raster : xarray.DataArray
        2D or 3D labeled data array.
    layer : float, optional
        For a 3D array, value along the z dimension.
    width : int, default=512
        Width of the output aggregate in pixels.
    height : int, default=None
        Height of the output aggregate in pixels.
        If not provided, height will be implied from `width`
        using aspect ratio of input raster.
    x_range : tuple of int, optional
        A tuple representing the bounds inclusive space ``[min, max]``
        along the x-axis.
    y_range : tuple of int, optional
        A tuple representing the bounds inclusive space ``[min, max]``
        along the y-axis.

    References
    ----------
        - https://datashader.org/_modules/datashader/core.html#Canvas
    """

    # get ranges
    if x_range is None:
        x_range = (
            raster.coords["x"].min().item(),
            raster.coords["x"].max().item()
        )
    if y_range is None:
        y_range = (
            raster.coords["y"].min().item(),
            raster.coords["y"].max().item()
        )

    if height is None:
        # set width and height
        height = height_implied_by_aspect_ratio(width, x_range, y_range)

    cvs = ds.Canvas(
        plot_width=width, plot_height=height, x_range=x_range, y_range=y_range
    )
    out = cvs.raster(raster, **kwargs)

    return out



import pytest
import xarray as xr
import numpy as np

import dask.array as da

from xrspatial import slope
from xrspatial.utils import doesnt_have_cuda


# Test Data -----------------------------------------------------------------

'''
Notes:
------
The `elevation` data was run through QGIS slope function to
get values to compare against.  Xarray-Spatial currently handles
edges by padding with nan which is different than QGIS but acknowledged
'''

elevation = np.asarray([
    [1432.6542, 1432.4764, 1432.4764, 1432.1207, 1431.9429, np.nan],
    [1432.6542, 1432.6542, 1432.4764, 1432.2986, 1432.1207, np.nan],
    [1432.832, 1432.6542, 1432.4764, 1432.2986, 1432.1207, np.nan],
    [1432.832, 1432.6542, 1432.4764, 1432.4764, 1432.1207, np.nan],
    [1432.832, 1432.6542, 1432.6542, 1432.4764, 1432.2986, np.nan],
    [1432.832, 1432.6542, 1432.6542, 1432.4764, 1432.2986, np.nan],
    [1432.832, 1432.832, 1432.6542, 1432.4764, 1432.4764, np.nan]],
    dtype=np.float32
)

qgis_slope = np.asarray(
    [[0.8052942, 0.742317, 1.1390567, 1.3716657, np.nan, np.nan],
     [0.74258685, 0.742317, 1.0500116, 1.2082565, np.nan, np.nan],
     [0.56964326, 0.9002944, 0.9002944, 1.0502871, np.nan, np.nan],
     [0.5095078, 0.9003686, 0.742317, 1.1390567, np.nan, np.nan],
     [0.6494868, 0.64938396, 0.5692523, 1.0500116, np.nan, np.nan],
     [0.80557066, 0.56964326, 0.64914393, 0.9002944, np.nan, np.nan],
     [0.6494868, 0.56964326, 0.8052942, 0.742317, np.nan, np.nan]],
    dtype=np.float32)

elevation2 = np.asarray([
    [np.nan, np.nan, np.nan, np.nan, np.nan, np.nan],
    [1584.8767, 1584.8767, 1585.0546, 1585.2324, 1585.2324, 1585.2324],
    [1585.0546, 1585.0546, 1585.2324, 1585.588, 1585.588, 1585.588],
    [1585.2324, 1585.4102, 1585.588, 1585.588, 1585.588, 1585.588],
    [1585.588, 1585.588, 1585.7659, 1585.7659, 1585.7659, 1585.7659],
    [1585.7659, 1585.9437, 1585.7659, 1585.7659, 1585.7659, 1585.7659],
    [1585.9437, 1585.9437, 1585.9437, 1585.7659, 1585.7659, 1585.7659]],
    dtype=np.float32
)


def test_slope_against_qgis():

    small_da = xr.DataArray(elevation, attrs={'res': (10.0, 10.0)})

    # slope by xrspatial
    xrspatial_slope = slope(small_da, name='slope_agg')

    # validate output attributes
    assert xrspatial_slope.dims == small_da.dims
    assert xrspatial_slope.attrs == small_da.attrs
    assert xrspatial_slope.shape == small_da.shape
    assert xrspatial_slope.name == 'slope_agg'
    for coord in small_da.coords:
        assert np.all(xrspatial_slope[coord] == small_da[coord])

    # validate output values
    # ignore border edges
    xrspatial_vals = xrspatial_slope.values[1:-1, 1:-1]
    qgis_vals = qgis_slope[1:-1, 1:-1]
    assert (np.isclose(xrspatial_vals, qgis_vals, equal_nan=True).all() | (
                np.isnan(xrspatial_vals) & np.isnan(qgis_vals))).all()


@pytest.mark.skipif(doesnt_have_cuda(),
                    reason="CUDA Device not Available")
def test_slope_against_qgis_gpu():

    import cupy

    small_da = xr.DataArray(elevation, attrs={'res': (10.0, 10.0)})
    small_da_cupy = xr.DataArray(cupy.asarray(elevation),
                                 attrs={'res': (10.0, 10.0)})
    xrspatial_slope = slope(small_da_cupy, name='slope_cupy')

    # validate output attributes
    assert xrspatial_slope.dims == small_da.dims
    assert xrspatial_slope.attrs == small_da.attrs
    assert xrspatial_slope.shape == small_da.shape
    for coord in small_da.coords:
        assert np.all(xrspatial_slope[coord] == small_da[coord])

    # validate output values
    # ignore border edges
    xrspatial_vals = xrspatial_slope.values[1:-1, 1:-1]
    qgis_vals = qgis_slope[1:-1, 1:-1]
    assert (np.isclose(xrspatial_vals, qgis_vals, equal_nan=True).all() | (
                np.isnan(xrspatial_vals) & np.isnan(qgis_vals))).all()


@pytest.mark.skipif(doesnt_have_cuda(),
                    reason="CUDA Device not Available")
def test_slope_gpu_equals_cpu():

    import cupy

    small_da = xr.DataArray(elevation2, attrs={'res': (10.0, 10.0)})
    cpu = slope(small_da, name='numpy_result')

    small_da_cupy = xr.DataArray(cupy.asarray(elevation2),
                                 attrs={'res': (10.0, 10.0)})
    gpu = slope(small_da_cupy, name='cupy_result')
    assert isinstance(gpu.data, cupy.ndarray)

    assert np.isclose(cpu, gpu, equal_nan=True).all()


@pytest.mark.skipif(doesnt_have_cuda(), reason="CUDA Device not Available")
def _dask_cupy_equals_numpy_cpu():

    # NOTE: Dask + GPU code paths don't currently work because of
    # dask casting cupy arrays to numpy arrays during
    # https://github.com/dask/dask/issues/4842

    import cupy

    cupy_data = cupy.asarray(elevation2)
    dask_cupy_data = da.from_array(cupy_data, chunks=(3, 3))

    small_da = xr.DataArray(elevation2, attrs={'res': (10.0, 10.0)})
    cpu = slope(small_da, name='numpy_result')

    small_dask_cupy = xr.DataArray(dask_cupy_data,
                                   attrs={'res': (10.0, 10.0)})
    gpu = slope(small_dask_cupy, name='cupy_result')

    assert np.isclose(cpu, gpu, equal_nan=True).all()


def test_slope_numpy_equals_dask():
    small_numpy_based_data_array = xr.DataArray(elevation2,
                                                attrs={'res': (10.0, 10.0)})
    small_das_based_data_array = xr.DataArray(da.from_array(elevation2,
                                              chunks=(3, 3)),
                                              attrs={'res': (10.0, 10.0)})

    numpy_slope = slope(small_numpy_based_data_array, name='numpy_slope')
    dask_slope = slope(small_das_based_data_array, name='dask_slope')
    assert isinstance(dask_slope.data, da.Array)

    dask_slope.data = dask_slope.data.compute()

    assert np.isclose(numpy_slope, dask_slope, equal_nan=True).all()


def test_slope_with_dask_array():

    import dask.array as da

    data = da.from_array(elevation, chunks=(3, 3))
    small_da = xr.DataArray(data, attrs={'res': (10.0, 10.0)})

    # slope by xrspatial
    xrspatial_slope = slope(small_da, name='slope_agg')
    xrspatial_slope.data = xrspatial_slope.data.compute()

    # validate output attributes
    assert xrspatial_slope.dims == small_da.dims
    assert xrspatial_slope.attrs == small_da.attrs
    assert xrspatial_slope.shape == small_da.shape
    assert xrspatial_slope.name == 'slope_agg'
    for coord in small_da.coords:
        assert np.all(xrspatial_slope[coord] == small_da[coord])

    # validate output values
    # ignore border edges
    xrspatial_vals = xrspatial_slope.values[1:-1, 1:-1]
    qgis_vals = qgis_slope[1:-1, 1:-1]
    assert (np.isclose(xrspatial_vals, qgis_vals, equal_nan=True).all() | (
                np.isnan(xrspatial_vals) & np.isnan(qgis_vals))).all()



from functools import partial

import re

import numpy as np
import dask.array as da
from xarray import DataArray

from numba import cuda, float32, prange, jit

from xrspatial.utils import has_cuda
from xrspatial.utils import cuda_args
from xrspatial.utils import get_dataarray_resolution

# 3rd-party
try:
    import cupy
except ImportError:
    class cupy(object):
        ndarray = False


DEFAULT_UNIT = 'meter'
METER = 1
FOOT = 0.3048
KILOMETER = 1000
MILE = 1609.344
UNITS = {'meter': METER, 'meters': METER, 'm': METER,
         'feet': FOOT, 'foot': FOOT, 'ft': FOOT,
         'miles': MILE, 'mls': MILE, 'ml': MILE,
         'kilometer': KILOMETER, 'kilometers': KILOMETER, 'km': KILOMETER}


def _is_numeric(s):
    try:
        float(s)
        return True
    except ValueError:
        return False


def _to_meters(d, unit):
    return d * UNITS[unit]


def _get_distance(distance_str):
    # return distance in meters

    # spit string into numbers and text
    splits = [x for x in re.split(r'(-?\d*\.?\d+)', distance_str) if x != '']
    if len(splits) not in [1, 2]:
        raise ValueError("Invalid distance.")

    unit = DEFAULT_UNIT

    if len(splits) == 2:
        unit = splits[1]

    number = splits[0]
    if not _is_numeric(number):
        raise ValueError("Distance should be a positive numeric value.\n")

    distance = float(number)
    if distance <= 0:
        raise ValueError("Distance should be a positive.\n")

    unit = unit.lower()
    unit = unit.replace(' ', '')
    if unit not in UNITS:
        raise ValueError(
            "Distance unit should be one of the following: \n"
            "meter (meter, meters, m),\n"
            "kilometer (kilometer, kilometers, km),\n"
            "foot (foot, feet, ft),\n"
            "mile (mile, miles, ml, mls)")

    # convert distance to meters
    meters = _to_meters(distance, unit)
    return meters


def calc_cellsize(raster):
    """
    Calculates cell size of an array based on its attributes.
    Default = meters. If lat-lon, units are converted to meters.

    Parameters
    ----------
    raster : xarray.DataArray
        2D array of input values.

    Returns
    -------
    cellsize : tuple
        Tuple of (cellsize_x, cellsize_y).
    cellsize_x : float
        Size of cells in x-direction.
    cellsize_y : float
        Size of cells in y-direction.

    Examples
    --------
    .. plot::
       :include-source:

        import datashader as ds
        import matplotlib.pyplot as plt
        from xrspatial import generate_terrain
        from xrspatial.convolution import calc_cellsize

        # Create Canvas
        W = 500
        H = 300
        cvs = ds.Canvas(plot_width = W,
                        plot_height = H,
                        x_range = (-20e6, 20e6),
                        y_range = (-20e6, 20e6))

        # Generate Example Terrain
        terrain_agg = generate_terrain(canvas = cvs)

        # Edit Attributes
        terrain_agg = terrain_agg.assign_attrs(
            {
                'Description': 'Example Terrain',
                'units': 'km',
                'Max Elevation': '4000',
            }
        )

        terrain_agg = terrain_agg.rename({'x': 'lon', 'y': 'lat'})
        terrain_agg = terrain_agg.rename('Elevation')

    .. sourcecode:: python

        >>> print(terrain_agg[200:203, 200:202])
        <xarray.DataArray 'Elevation' (lat: 3, lon: 2)>
        array([[1264.02249454, 1261.94748873],
               [1285.37061171, 1282.48046696],
               [1306.02305679, 1303.40657515]])
        Coordinates:
          * lon      (lon) float64 -3.96e+06 -3.88e+06
          * lat      (lat) float64 6.733e+06 6.867e+06 7e+06
        Attributes:
            res:            1
            Description:    Example Terrain
            Max Elevation:  3000
            units:          km

    .. sourcecode:: python

        >>> # Calculate Cellsize
        >>> cellsize = calc_cellsize(terrain_agg)
        >>> print(cellsize)
        (1, 1)
    """
    if 'unit' in raster.attrs:
        unit = raster.attrs['unit']
    else:
        unit = DEFAULT_UNIT

    cellsize_x, cellsize_y = get_dataarray_resolution(raster)
    cellsize_x = _to_meters(cellsize_x, unit)
    cellsize_y = _to_meters(cellsize_y, unit)

    # When converting from lnglat_to_meters, could have negative cellsize in y
    return cellsize_x, np.abs(cellsize_y)


def _ellipse_kernel(half_w, half_h):
    # x values of interest
    x = np.linspace(-half_w, half_w, 2 * half_w + 1)
    # y values of interest, as a "column" array
    y = np.linspace(-half_h, half_h, 2 * half_h + 1)[:, None]

    # True for points inside the ellipse
    # (x / a)^2 + (y / b)^2 <= 1, avoid division to avoid rounding issue
    ellipse = (x * half_h) ** 2 + (y * half_w) ** 2 <= (half_w * half_h) ** 2
    return ellipse.astype(float)


def circle_kernel(cellsize_x, cellsize_y, radius):
    """
    Generates a circular kernel of a given cellsize and radius.

    Parameters
    ----------
    cellsize_x : int
        Cell size of output kernel in x-direction.
    cellsize_y : int
        Cell size of output kernel in y-direction.
    radius : int
        Radius of output kernel.

    Returns
    -------
    kernel : NumPy Array of float values
        2D array where values of 1 indicate the kernel.

    Examples
    --------
    .. sourcecode:: python

        >>> import xarray as xr
        >>> from xrspatial.convolution import circle_kernel

        >>> # Create Kernel
        >>> kernel = circle_kernel(1, 1, 3)
        >>> print(kernel)
        [[0. 0. 0. 1. 0. 0. 0.]
        [0. 1. 1. 1. 1. 1. 0.]
        [0. 1. 1. 1. 1. 1. 0.]
        [1. 1. 1. 1. 1. 1. 1.]
        [0. 1. 1. 1. 1. 1. 0.]
        [0. 1. 1. 1. 1. 1. 0.]
        [0. 0. 0. 1. 0. 0. 0.]]

        >>> kernel = circle_kernel(1, 2, 3)
        >>> print(kernel)
        [[0. 0. 0. 1. 0. 0. 0.]
         [1. 1. 1. 1. 1. 1. 1.]
         [0. 0. 0. 1. 0. 0. 0.]]
    """
    # validate radius, convert radius to meters
    r = _get_distance(str(radius))

    kernel_half_w = int(r / cellsize_x)
    kernel_half_h = int(r / cellsize_y)

    kernel = _ellipse_kernel(kernel_half_w, kernel_half_h)
    return kernel


def annulus_kernel(cellsize_x, cellsize_y, outer_radius, inner_radius):
    """
    Generates a annulus (ring-shaped) kernel of a given cellsize and radius.

    Parameters
    ----------
    cellsize_x : int
        Cell size of output kernel in x direction.
    cellsize_y : int
        Cell size of output kernel in y direction.
    outer_radius : int
        Outer ring radius of output kernel.
    inner_radius : int
        Inner circle radius of output kernel.

    Returns
    -------
    kernel : NumPy Array of float values.
        2D array of 0s and 1s where values of 1 indicate the kernel.

    Examples
    --------
    .. sourcecode:: python

        >>> import xarray as xr
        >>> from xrspatial.convolution import annulus_kernel

        >>> # Create Kernel
        >>> kernel = annulus_kernel(1, 1, 3, 1)
        >>> print(kernel)
        [[0., 0., 0., 1., 0., 0., 0.],
         [0., 1., 1., 1., 1., 1., 0.],
         [0., 1., 1., 0., 1., 1., 0.],
         [1., 1., 0., 0., 0., 1., 1.],
         [0., 1., 1., 0., 1., 1., 0.],
         [0., 1., 1., 1., 1., 1., 0.],
         [0., 0., 0., 1., 0., 0., 0.]]

        >>> kernel = annulus_kernel(1, 2, 5, 2)
        >>> print(kernel)
        [[0., 0., 0., 0., 0., 1., 0., 0., 0., 0., 0.],
         [0., 1., 1., 1., 1., 0., 1., 1., 1., 1., 0.],
         [1., 1., 1., 0., 0., 0., 0., 0., 1., 1., 1.],
         [0., 1., 1., 1., 1., 0., 1., 1., 1., 1., 0.],
         [0., 0., 0., 0., 0., 1., 0., 0., 0., 0., 0.]])
    """
    # Get the two circular kernels for the annulus
    kernel_outer = circle_kernel(cellsize_x, cellsize_y, outer_radius)
    kernel_inner = circle_kernel(cellsize_x, cellsize_y, inner_radius)

    # Need to pad kernel_inner to get it the same shape and centered
    # in kernel_outer
    pad_vals = np.array(kernel_outer.shape) - np.array(kernel_inner.shape)
    pad_kernel = np.pad(kernel_inner,
                        # Pad ((before_rows, after_rows),
                        #      (before_cols, after_cols))
                        pad_width=((pad_vals[0] // 2, pad_vals[0] // 2),
                                   (pad_vals[1] // 2, pad_vals[1] // 2)),
                        mode='constant',
                        constant_values=0)
    # Get annulus by subtracting inner from outer
    kernel = kernel_outer - pad_kernel
    return kernel


def custom_kernel(kernel):
    """
    Validates a custom kernel. If the kernel is valid, returns itself.
    """

    if not isinstance(kernel, np.ndarray):
        raise ValueError(
            "Received a custom kernel that is not a Numpy array.",
            "The kernel received was of type {} and needs to be "
            "of type `ndarray`".format(type(kernel))
        )
    else:
        rows, cols = kernel.shape

    if (rows % 2 == 0 or cols % 2 == 0):
        raise ValueError(
            "Received custom kernel with improper dimensions.",
            "A custom kernel needs to have an odd shape, the supplied kernel "
            "has {} rows and {} columns.".format(rows, cols)
        )
    return kernel


@jit(nopython=True, nogil=True, parallel=True)
def _convolve_2d_numpy(data, kernel):
    """
    Apply kernel to data image.
    """
    # TODO: handle nan

    nx = data.shape[0]
    ny = data.shape[1]
    nkx = kernel.shape[0]
    nky = kernel.shape[1]
    wkx = nkx // 2
    wky = nky // 2

    out = np.zeros(data.shape, dtype=float32)
    out[:, :] = np.nan
    for i in prange(wkx, nx-wkx):
        iimin = max(i - wkx, 0)
        iimax = min(i + wkx + 1, nx)
        for j in prange(wky, ny-wky):
            jjmin = max(j - wky, 0)
            jjmax = min(j + wky + 1, ny)
            num = 0.0
            for ii in range(iimin, iimax, 1):
                iii = wkx + ii - i
                for jj in range(jjmin, jjmax, 1):
                    jjj = wky + jj - j
                    num += kernel[iii, jjj] * data[ii, jj]
            out[i, j] = num

    return out


def _convolve_2d_dask_numpy(data, kernel):
    pad_h = kernel.shape[0] // 2
    pad_w = kernel.shape[1] // 2
    _func = partial(_convolve_2d_numpy, kernel=kernel)
    out = data.map_overlap(_func,
                           depth=(pad_h, pad_w),
                           boundary=np.nan,
                           meta=np.array(()))
    return out


# https://www.vincent-lunot.com/post/an-introduction-to-cuda-in-python-part-3/
@cuda.jit
def _convolve_2d_cuda(data, kernel, out):
    # expect a 2D grid and 2D blocks,
    # a kernel with odd numbers of rows and columns, (-1-)
    # a grayscale image

    # (-2-) 2D coordinates of the current thread:
    i, j = cuda.grid(2)

    # To compute the out at coordinates (i, j), we need to use delta_rows rows
    # of the array before and after the i_th row, as well as delta_cols columns
    # of the array before and after the j_th column:
    delta_rows = kernel.shape[0] // 2
    delta_cols = kernel.shape[1] // 2

    data_rows, data_cols = data.shape
    # (-3-) if the thread coordinates are outside of the data image,
    # we ignore the thread
    # currently, if the thread coordinates are in the edges,
    # we ignore the thread
    if i < delta_rows or i >= data_rows - delta_rows or \
            j < delta_cols or j >= data_cols - delta_cols:
        return

    # The out at coordinates (i, j) is equal to
    # sum_{k, h} kernel[k, h] * data[i - k + delta_rows, j - h + delta_cols]
    # with k and h going through the whole kernel array:
    s = 0
    for k in range(kernel.shape[0]):
        for h in range(kernel.shape[1]):
            i_k = i - k + delta_rows
            j_h = j - h + delta_cols
            # (-4-) Check if (i_k, j_h) coordinates are inside the array:
            if (i_k >= 0) and (i_k < data_rows) and \
                    (j_h >= 0) and (j_h < data_cols):
                s += kernel[k, h] * data[i_k, j_h]
    out[i, j] = s


def _convolve_2d_cupy(data, kernel):
    out = cupy.empty(data.shape, dtype='f4')
    out[:, :] = cupy.nan
    griddim, blockdim = cuda_args(data.shape)
    _convolve_2d_cuda[griddim, blockdim](data, kernel, cupy.asarray(out))
    return out


def _convolve_2d_dask_cupy(data, kernel):
    msg = 'Upstream bug in dask prevents cupy backed arrays'
    raise NotImplementedError(msg)


def convolve_2d(data, kernel):
    """
    Calculates, for all inner cells of an array, the 2D convolution of
    each cell via Numba. To account for edge cells, a pad can be added
    to the image array. Convolution is frequently used for image
    processing, such as smoothing, sharpening, and edge detection of
    images by eliminating spurious data or enhancing features in the
    data.

    Parameters
    ----------
    image : xarray.DataArray
        2D array of values to processed and padded.
    kernel : array-like object
        Impulse kernel, determines area to apply impulse function for
        each cell.
    pad : bool, default=True
        To compute edges set to True.
    use-cuda : bool, default=True
        For parallel computing set to True.

    Returns
    -------
    convolve_agg : numpy.ndarray
        2D array representation of the impulse function.

    Examples
    --------
    .. plot::
       :include-source:

        import numpy as np
        import xarray as xr
        from xrspatial import focal
        from xrspatial.convolution import convolve_2d

        # Create Data Array
        agg = xr.DataArray(np.array([[0, 0, 0, 0, 0, 0, 0],
                                     [0, 0, 2, 4, 0, 8, 0],
                                     [0, 2, 2, 4, 6, 8, 0],
                                     [0, 4, 4, 4, 6, 8, 0],
                                     [0, 6, 6, 6, 6, 8, 0],
                                     [0, 8, 8, 8, 8, 8, 0],
                                     [0, 0, 0, 0, 0, 0, 0]]),
                            dims = ["lat", "lon"],
                            attrs = dict(res = 1))
        height, width = agg.shape
        _lon = np.linspace(0, width - 1, width)
        _lat = np.linspace(0, height - 1, height)
        agg["lon"] = _lon
        agg["lat"] = _lat

        # Create Kernel
        kernel = focal.circle_kernel(1, 1, 1)

        # Create Convolution Data Array
        convolve_agg = convolve_2d(image = agg, kernel = kernel)

    .. sourcecode:: python

        >>> print(convolve_agg)
        [[ 0.  0.  4.  8.  0. 16.  0.]
        [ 0.  4.  8. 10. 18. 16. 16.]
        [ 4.  8. 14. 20. 24. 30. 16.]
        [ 8. 16. 20. 24. 30. 30. 16.]
        [12. 24. 30. 30. 34. 30. 16.]
        [16. 22. 30. 30. 30. 24. 16.]
        [ 0. 16. 16. 16. 16. 16.  0.]]
    """
    # numpy case
    if isinstance(data, np.ndarray):
        out = _convolve_2d_numpy(data, kernel)

    # cupy case
    elif has_cuda() and isinstance(data, cupy.ndarray):
        out = _convolve_2d_cupy(data, kernel)

    # dask + cupy case
    elif has_cuda() and isinstance(data, da.Array) and \
            type(data._meta).__module__.split('.')[0] == 'cupy':
        out = _convolve_2d_dask_cupy(data, kernel)

    # dask + numpy case
    elif isinstance(data, da.Array):
        out = _convolve_2d_dask_numpy(data, kernel)

    else:
        raise TypeError('Unsupported Array Type: {}'.format(type(data)))

    return out


def convolution_2d(agg, kernel):
    """
    Calculates, for all inner cells of an array, the 2D convolution of
    each cell via Numba. To account for edge cells, a pad can be added
    to the image array. Convolution is frequently used for image
    processing, such as smoothing, sharpening, and edge detection of
    images by eliminating spurious data or enhancing features in the
    data.

    Parameters
    ----------
    agg : xarray.DataArray
        2D array of values to processed and padded.
    kernel : array-like object
        Impulse kernel, determines area to apply impulse function for
        each cell.

    Returns
    -------
    convolve_agg : xarray.DataArray
        2D array representation of the impulse function.
    """

    # wrapper of convolve_2d
    out = convolve_2d(agg.data, kernel)

    return DataArray(out,
                     coords=agg.coords,
                     dims=agg.dims,
                     attrs=agg.attrs)



import datashader as ds

from xrspatial import generate_terrain

W = 25
H = 30

X_RANGE = (0, 500)
Y_RANGE = (0, 500)


def test_generate_terrain():
    csv = ds.Canvas(x_range=X_RANGE, y_range=Y_RANGE,
                    plot_width=W, plot_height=H)
    terrain = generate_terrain(canvas=csv)
    assert terrain is not None



import pytest

import xarray as xr
import numpy as np

import dask.array as da

from xrspatial.utils import doesnt_have_cuda, is_cupy_backed
from xrspatial import equal_interval
from xrspatial import natural_breaks
from xrspatial import quantile
from xrspatial import reclassify

elevation = np.array([
    [1.,  2.,  3.,  4., np.nan],
    [5.,  6.,  7.,  8.,  9.],
    [10., 11., 12., 13., 14.],
    [15., 16., 17., 18., np.inf],
])

numpy_agg = xr.DataArray(elevation, attrs={'res': (10.0, 10.0)})
dask_numpy_agg = xr.DataArray(da.from_array(elevation, chunks=(3, 3)),
                              attrs={'res': (10.0, 10.0)})


def test_reclassify_cpu():
    bins = [10, 20]
    new_values = [1, 2]

    # numpy
    numpy_reclassify = reclassify(numpy_agg, bins=bins, new_values=new_values,
                                  name='numpy_reclassify')
    # ignore nans
    unique_elements = np.unique(
        numpy_reclassify.data[np.isfinite(numpy_reclassify.data)]
    )
    assert len(unique_elements) == 2

    # dask + numpy
    dask_reclassify = reclassify(dask_numpy_agg, bins=bins,
                                 new_values=new_values, name='dask_reclassify')
    assert isinstance(dask_reclassify.data, da.Array)

    dask_reclassify.data = dask_reclassify.data.compute()
    assert np.isclose(numpy_reclassify, dask_reclassify, equal_nan=True).all()


@pytest.mark.skipif(doesnt_have_cuda(), reason="CUDA Device not Available")
def test_reclassify_cpu_equals_gpu():

    import cupy

    bins = [10, 20, 30]
    new_values = [1, 2, 3]

    # vanilla numpy version
    cpu = reclassify(numpy_agg,
                     name='numpy_result',
                     bins=bins,
                     new_values=new_values)

    # cupy
    cupy_agg = xr.DataArray(cupy.asarray(elevation),
                            attrs={'res': (10.0, 10.0)})
    gpu = reclassify(cupy_agg,
                     name='cupy_result',
                     bins=bins,
                     new_values=new_values)
    assert isinstance(gpu.data, cupy.ndarray)
    assert np.isclose(cpu, gpu, equal_nan=True).all()

    # dask + cupy
    dask_cupy_agg = xr.DataArray(cupy.asarray(elevation),
                                 attrs={'res': (10.0, 10.0)})
    dask_cupy_agg.data = da.from_array(dask_cupy_agg.data, chunks=(3, 3))
    dask_gpu = reclassify(dask_cupy_agg, name='dask_cupy_result',
                          bins=bins, new_values=new_values)
    assert isinstance(dask_gpu.data, da.Array) and is_cupy_backed(dask_gpu)

    dask_gpu.data = dask_gpu.data.compute()
    assert np.isclose(cpu, dask_gpu, equal_nan=True).all()


def test_quantile_cpu():
    k = 5

    # numpy
    numpy_quantile = quantile(numpy_agg, k=k)
    unique_elements = np.unique(
        numpy_quantile.data[np.isfinite(numpy_quantile.data)]
    )
    assert isinstance(numpy_quantile.data, np.ndarray)
    assert len(unique_elements) == k

    # dask + numpy
    dask_quantile = quantile(dask_numpy_agg, k=k)
    assert isinstance(dask_quantile.data, da.Array)

    #     Note that dask's percentile algorithm is
    #     approximate, while numpy's is exact.
    #     This may cause some differences between
    #     results of vanilla numpy and
    #     dask version of the input agg.
    #     https://github.com/dask/dask/issues/3099
    #     This assertion may fail
    # dask_quantile = dask_quantile.compute()
    # assert np.isclose(numpy_quantile, dask_quantile, equal_nan=True).all()


@pytest.mark.skipif(doesnt_have_cuda(), reason="CUDA Device not Available")
def test_quantile_cpu_equals_gpu():

    import cupy

    k = 5

    # vanilla numpy version
    cpu = quantile(numpy_agg, k=k, name='numpy_result')

    # cupy
    cupy_agg = xr.DataArray(cupy.asarray(elevation),
                            attrs={'res': (10.0, 10.0)})
    gpu = quantile(cupy_agg, k=k, name='cupy_result')

    assert isinstance(gpu.data, cupy.ndarray)
    assert np.isclose(cpu, gpu, equal_nan=True).all()


def test_natural_breaks_cpu():
    k = 5

    # vanilla numpy
    numpy_natural_breaks = natural_breaks(numpy_agg, k=k)
    # shape and other attributes remain the same
    assert numpy_agg.shape == numpy_natural_breaks.shape
    assert numpy_agg.dims == numpy_natural_breaks.dims
    assert numpy_agg.attrs == numpy_natural_breaks.attrs
    for coord in numpy_agg.coords:
        assert np.all(numpy_agg[coord] == numpy_natural_breaks[coord])

    unique_elements = np.unique(
        numpy_natural_breaks.data[np.isfinite(numpy_natural_breaks.data)]
    )
    assert len(unique_elements) == k


@pytest.mark.skipif(doesnt_have_cuda(), reason="CUDA Device not Available")
def test_natural_breaks_cpu_equals_gpu():

    import cupy

    k = 5

    # vanilla numpy version
    cpu = natural_breaks(numpy_agg, k=k, name='numpy_result')

    # cupy
    cupy_agg = xr.DataArray(cupy.asarray(elevation),
                            attrs={'res': (10.0, 10.0)})
    gpu = natural_breaks(cupy_agg, k=k, name='cupy_result')

    assert isinstance(gpu.data, cupy.ndarray)
    assert np.isclose(cpu, gpu, equal_nan=True).all()


def test_equal_interval_cpu():
    k = 5
    # numpy
    numpy_ei = equal_interval(numpy_agg, k=5)

    unique_elements = np.unique(numpy_ei.data[da.isfinite(numpy_ei.data)])
    assert isinstance(numpy_ei.data, np.ndarray)
    assert len(unique_elements) == k

    # dask + numpy
    dask_ei = equal_interval(dask_numpy_agg, k=k, name='dask_reclassify')
    assert isinstance(dask_ei.data, da.Array)

    dask_ei.data = dask_ei.data.compute()
    assert np.isclose(numpy_ei, dask_ei, equal_nan=True).all()


@pytest.mark.skipif(doesnt_have_cuda(), reason="CUDA Device not Available")
def test_equal_interval_cpu_equals_gpu():

    import cupy

    k = 5

    # numpy
    cpu = equal_interval(numpy_agg, k=k)

    # cupy
    cupy_agg = xr.DataArray(cupy.asarray(elevation),
                            attrs={'res': (10.0, 10.0)})
    gpu = equal_interval(cupy_agg, k=k)
    assert isinstance(gpu.data, cupy.ndarray)

    assert np.isclose(cpu, gpu, equal_nan=True).all()



import numpy as np

import xarray as xr
from xarray import DataArray

from xrspatial.utils import ngjit


# TODO: change parameters to take agg instead of height / width
def perlin(width: int,
           height: int,
           freq: tuple = (1, 1),
           seed: int = 5) -> xr.DataArray:
    """
    Generate perlin noise aggregate.

    Parameters
    ----------
    width : int
        Width of output aggregate array.
    height : int
        Height of output aggregate array.
    freq : tuple, default=(1,1)
        (x, y) frequency multipliers.
    seed : int, default=5
        Seed for random number generator.

    Returns
    -------
    perlin_agg : xarray.DataArray
        2D array of perlin noise values.

    References
    ----------
        - Paul Panzer: https://stackoverflow.com/questions/42147776/producing-2d-perlin-noise-with-numpy # noqa
        - ICA: http://www.mountaincartography.org/mt_hood/pdfs/nighbert_bump1.pdf # noqa

    Examples
    --------
    .. plot::
       :include-source:

        import matplotlib.pyplot as plt
        from xrspatial import perlin

        # Generate Perlin Noise Aggregate
        perlin_default = perlin(width = 500, height = 300)

        # With Increased x Frequency
        perlin_high_x_freq = perlin(width = 500, height = 300, freq = (5, 1))

        # With Increased y Frequency
        perlin_high_y_freq = perlin(width = 500, height = 300, freq = (1, 5))

        # With a Different Seed
        perlin_seed_1 = perlin(width = 500, height = 300, seed = 1)

        # Plot Default Perlin
        perlin_default.plot(cmap = 'inferno', aspect = 2, size = 4)
        plt.title("Default")

        # Plot High x Frequency
        perlin_high_x_freq.plot(cmap = 'inferno', aspect = 2, size = 4)
        plt.title("High x Frequency")

        # Plot High y Frequency
        perlin_high_y_freq.plot(cmap = 'inferno', aspect = 2, size = 4)
        plt.title("High y Frequency")

        # Plot Seed = 1
        perlin_seed_1.plot(cmap = 'inferno', aspect = 2, size = 4)
        plt.title("Seed = 1")

    .. sourcecode:: python

        >>> print(perlin_default[200:203, 200: 202])
        <xarray.DataArray (y: 3, x: 2)>
        array([[0.56800979, 0.56477393],
               [0.56651744, 0.56331014],
               [0.56499184, 0.56181344]])
        Dimensions without coordinates: y, x
        Attributes:
            res:      1

        >>> print(perlin_high_x_freq[200:203, 200: 202])
        <xarray.DataArray (y: 3, x: 2)>
        array([[0.5       , 0.48999444],
               [0.5       , 0.48999434],
               [0.5       , 0.48999425]])
        Dimensions without coordinates: y, x
        Attributes:
            res:      1

        >>> print(perlin_high_y_freq[200:203, 200: 202])
        <xarray.DataArray (y: 3, x: 2)>
        array([[0.31872961, 0.31756859],
               [0.2999256 , 0.2988189 ],
               [0.28085118, 0.27979834]])
        Dimensions without coordinates: y, x
        Attributes:
            res:      1

        >>> print(perlin_seed_1[200:203, 200: 202])
        <xarray.DataArray (y: 3, x: 2)>
        array([[0.12991498, 0.12984185],
               [0.13451158, 0.13441514],
               [0.13916956, 0.1390495 ]])
        Dimensions without coordinates: y, x
        Attributes:
            res:      1
    """
    linx = range(width)
    liny = range(height)
    linx = np.linspace(0, 1, width, endpoint=False)
    liny = np.linspace(0, 1, height, endpoint=False)
    x, y = np.meshgrid(linx, liny)
    data = _perlin(x * freq[0], y * freq[1], seed=seed)
    data = (data - np.min(data))/np.ptp(data)
    return DataArray(data, dims=['y', 'x'], attrs=dict(res=1))


@ngjit
def _lerp(a, b, x):
    return a + x * (b-a)


@ngjit
def _fade(t):
    return 6 * t**5 - 15 * t**4 + 10 * t**3


@ngjit
def _gradient(h, x, y):
    vectors = np.array([[0, 1], [0, -1], [1, 0], [-1, 0]])
    dim_ = h.shape
    out = np.zeros(dim_)
    for j in range(dim_[1]):
        for i in range(dim_[0]):
            f = np.mod(h[i, j], 4)
            g = vectors[f]
            out[i, j] = g[0] * x[i, j] + g[1] * y[i, j]
    return out


def _perlin(x, y, seed=0):
    np.random.seed(seed)
    p = np.arange(2**20, dtype=int)
    np.random.shuffle(p)
    p = np.stack([p, p]).flatten()

    # coordinates of the top-left
    xi = x.astype(int)
    yi = y.astype(int)

    # internal coordinates
    xf = x - xi
    yf = y - yi

    # fade factors
    u = _fade(xf)
    v = _fade(yf)

    # noise components
    n00 = _gradient(p[p[xi]+yi], xf, yf)
    n01 = _gradient(p[p[xi]+yi+1], xf, yf-1)
    n11 = _gradient(p[p[xi+1]+yi+1], xf-1, yf-1)
    n10 = _gradient(p[p[xi+1]+yi], xf-1, yf)

    # combine noises
    x1 = _lerp(n00, n10, u)
    x2 = _lerp(n01, n11, u)
    a = _lerp(x1, x2, v)
    return a



import os
import sys
import shutil
from setuptools import setup


# build dependencies
import pyct.build
import param

# dependencies

# datashader first, then pyct unless pyct version compatible with ds
# is specified
# spatialpandas may not be required in final pharmacy_desert version
# pyct may not be required after pyctdev is released
install_requires = [
    'dask',
    'datashader',
    'numba',
    'numpy>=1.7,<=1.20',
    'pandas',
    'pillow',
    'requests',
    'xarray',
    'pyct <=0.4.6',
    'param >=1.6.1',
    'distributed >=2021.03.0',
    'spatialpandas',
]

examples = [
]

# Additional tests dependencies and examples_extra may be needed in the future
extras_require = {
    'tests': [
        'pytest',
        'noise >=1.2.2',
    ],
    'examples': examples,
}

# additional doc dependencies may be needed
extras_require['doc'] = extras_require['examples'] + ['numpydoc']

extras_require['all'] = sorted(set(sum(extras_require.values(), [])))

version = param.version.get_setup_version(__file__, 'xarray-spatial',
                                          pkgname='xrspatial',
                                          archive_commit="$Format:%h$")

if 'sdist' in sys.argv and 'bdist_wheel' in sys.argv:
    try:
        version_test = version.split('.post')[1]
        version = version.split('.post')[0]
    except IndexError:
        version = version.split('+')[0]
    if version is None:
        sys.exit('invalid version')


# metadata for setuptools

setup_args = dict(
    name='xarray-spatial',
    version=version,
    description='xarray-based spatial analysis tools',
    install_requires=install_requires,
    extras_require=extras_require,
    tests_require=extras_require['tests'],
    zip_safe=False,
    classifiers=['Programming Language :: Python :: 3',
                 'License :: OSI Approved :: MIT License',
                 'Operating System :: OS Independent'],
    packages=['xrspatial',
              'xrspatial.tests'
              ],
    include_package_data=True,
    entry_points={
        'console_scripts': [
            'xrspatial = xrspatial.__main__:main'
        ]
    },
)


if __name__ == '__main__':
    example_path = os.path.join(os.path.dirname(os.path.abspath(__file__)),
                                'xrspatial', 'examples')
    if 'develop' not in sys.argv:
        pyct.build.examples(example_path, __file__, force=True)
    setup(**setup_args)

    if os.path.isdir(example_path):
        shutil.rmtree(example_path)



from math import sqrt
from typing import Optional, Callable, Union, Dict, List

import dask.array as da
import dask.dataframe as dd
import numpy as np
import pandas as pd
import xarray as xr
from xarray import DataArray

from xrspatial.utils import ngjit


def _stats_count(data):
    if isinstance(data, np.ndarray):
        # numpy case
        stats_count = np.ma.count(data)
    else:
        # dask case
        stats_count = data.size - da.ma.getmaskarray(data).sum()
    return stats_count


_DEFAULT_STATS = dict(
    mean=lambda z: z.mean(),
    max=lambda z: z.max(),
    min=lambda z: z.min(),
    sum=lambda z: z.sum(),
    std=lambda z: z.std(),
    var=lambda z: z.var(),
    count=lambda z: _stats_count(z),
)


def _to_int(numeric_value):
    # convert an integer in float type to integer type
    # if not an integer, return the value itself
    if float(numeric_value).is_integer():
        return int(numeric_value)
    return numeric_value


def _zone_cat_data(
    zones,
    values,
    zone_id,
    nodata_values,
    cat=None,
    cat_id=None
):

    # array backend
    if isinstance(zones.data, np.ndarray):
        array_module = np
    elif isinstance(zones.data, da.Array):
        array_module = da

    if len(values.shape) == 2:
        # 2D case
        conditions = (
            (zones.data != zone_id)
            | ~np.isfinite(values.data)  # mask out nan, inf
            | (values.data == nodata_values)  # mask out nodata_values
        )

        if cat is not None:
            conditions |= values.data != cat

        zone_cat_data = array_module.ma.masked_where(conditions, values.data)

    else:
        # 3D case
        cat_data = values[cat_id].data
        cat_masked_data = array_module.ma.masked_invalid(cat_data)
        zone_cat_data = array_module.ma.masked_where(
            (
                (zones.data != zone_id)
                | (cat_data == nodata_values)
            ),
            cat_masked_data
        )
    return zone_cat_data


def _stats(
    zones: xr.DataArray,
    values: xr.DataArray,
    unique_zones: List[int],
    stats_funcs: List,
    nodata_values: Union[int, float],
) -> Dict:

    stats_dict = {}
    # zone column
    stats_dict["zone"] = unique_zones
    # stats columns
    for stats in stats_funcs:
        stats_dict[stats] = []

    for zone_id in unique_zones:
        # get zone values
        zone_values = _zone_cat_data(zones, values, zone_id, nodata_values)
        for stats in stats_funcs:
            stats_func = stats_funcs.get(stats)
            if not callable(stats_func):
                raise ValueError(stats)
            stats_dict[stats].append(stats_func(zone_values))

    unique_zones = list(map(_to_int, unique_zones))
    stats_dict["zone"] = unique_zones

    return stats_dict


def _stats_numpy(
    zones: xr.DataArray,
    values: xr.DataArray,
    zone_ids: List[Union[int, float]],
    stats_funcs: Dict,
    nodata_zones: Union[int, float],
    nodata_values: Union[int, float],
) -> pd.DataFrame:

    if zone_ids is None:
        # no zone_ids provided, find ids for all zones
        # do not consider zone with nodata values
        unique_zones = np.unique(zones.data[np.isfinite(zones.data)])
        unique_zones = sorted(list(set(unique_zones) - set([nodata_zones])))
    else:
        unique_zones = np.array(zone_ids)

    stats_dict = _stats(
        zones,
        values,
        unique_zones,
        stats_funcs,
        nodata_values
    )

    stats_df = pd.DataFrame(stats_dict)
    stats_df.set_index("zone")

    return stats_df


def _stats_dask(
    zones: xr.DataArray,
    values: xr.DataArray,
    zone_ids: List[Union[int, float]],
    stats_funcs: Dict,
    nodata_zones: Union[int, float],
    nodata_values: Union[int, float],
) -> pd.DataFrame:

    if zone_ids is None:
        # no zone_ids provided, find ids for all zones
        # precompute unique zones
        unique_zones = da.unique(zones.data[da.isfinite(zones.data)]).compute()
        # do not consider zone with nodata values
        unique_zones = sorted(list(set(unique_zones) - set([nodata_zones])))
    else:
        unique_zones = np.array(zone_ids)

    stats_dict = _stats(
        zones,
        values,
        unique_zones,
        stats_funcs,
        nodata_values
    )

    stats_dict = {
        stats: da.stack(zonal_stats, axis=0)
        for stats, zonal_stats in stats_dict.items()
    }

    # generate dask dataframe
    stats_df = dd.concat(
        [dd.from_dask_array(stats) for stats in stats_dict.values()], axis=1
    )
    # name columns
    stats_df.columns = stats_dict.keys()
    stats_df.set_index("zone")

    return stats_df


def stats(
    zones: xr.DataArray,
    values: xr.DataArray,
    zone_ids: Optional[List[Union[int, float]]] = None,
    stats_funcs: Union[Dict, List] = [
        "mean",
        "max",
        "min",
        "sum",
        "std",
        "var",
        "count",
    ],
    nodata_zones: Optional[Union[int, float]] = None,
    nodata_values: Union[int, float] = None,
) -> Union[pd.DataFrame, dd.DataFrame]:
    """
    Calculate summary statistics for each zone defined by a zone
    dataset, based on values aggregate.

    A single output value is computed for every zone in the input zone
    dataset.

    This function currently supports numpy backed, and dask with numpy backed
    xarray DataArray.

    Parameters
    ----------
    zones : xr.DataArray
        zones is a 2D xarray DataArray of numeric values.
        A zone is all the cells in a raster that have the same value,
        whether or not they are contiguous. The input `zones` raster defines
        the shape, values, and locations of the zones. An integer field
        in the input `zones` DataArray defines a zone.

    values : xr.DataArray
        values is a 2D xarray DataArray of numeric values (integers or floats).
        The input `values` raster contains the input values used in
        calculating the output statistic for each zone.

    zone_ids : list of ints, or floats
        List of zones to be included in calculation. If no zone_ids provided,
        all zones will be used.

    stats_funcs : dict, or list of strings, default=['mean', 'max', 'min',
        'sum', 'std', 'var', 'count'])
        The statistics to calculate for each zone. If a list, possible
        choices are subsets of the default options.
        In the dictionary case, all of its values must be
        callable. Function takes only one argument that is the `values` raster.
        The key become the column name in the output DataFrame.

    nodata_zones: int, float, default=None
        Nodata value in `zones` raster.
        Cells with `nodata_zones` do not belong to any zone,
        and thus excluded from calculation.

    nodata_values: int, float, default=None
        Nodata value in `values` raster.
        Cells with `nodata_values` do not belong to any zone,
        and thus excluded from calculation.

    Returns
    -------
    stats_df : Union[pandas.DataFrame, dask.dataframe.DataFrame]
        A pandas DataFrame, or a dask DataFrame where each column
        is a statistic and each row is a zone with zone id.

    Examples
    --------
    .. plot::
       :include-source:

        import numpy as np
        import xarray as xr
        from xrspatial.zonal import stats

        height, width = 10, 10
        # Values raster
        values = xr.DataArray(np.arange(height * width).reshape(height, width))
        # Zones raster
        zones = xr.DataArray(np.zeros(height * width).reshape(height, width))
        zones[:5, :5] = 0
        zones[:5, 5:] = 10
        zones[5:, :5] = 20
        zones[5:, 5:] = 30

    .. sourcecode:: python

        >>> # Calculate Stats
        >>> stats_df = stats(zones=zones, values=values)
        >>> print(stats_df)
            zone  mean  max  min   sum       std    var  count
        0   0    22.0   44    0   550  14.21267  202.0     25
        1  10    27.0   49    5   675  14.21267  202.0     25
        2  20    72.0   94   50  1800  14.21267  202.0     25
        3  30    77.0   99   55  1925  14.21267  202.0     25

        >>> # Custom Stats
        >>> custom_stats ={'double_sum': lambda val: val.sum()*2}
        >>> custom_stats_df = stats(zones=zones,
                                    values=values,
                                    stats_funcs=custom_stats)
        >>> print(custom_stats_df)
            zone  double_sum
        0   0     1100
        1  10     1350
        2  20     3600
        3  30     3850

        >>> # Calculate Stats with dask backed xarray DataArrays
        >>> dask_stats_df = stats(zones=dask_zones, values=dask_values)
        >>> print(type(dask_stats_df))
        <class 'dask.dataframe.core.DataFrame'>
        >>> print(dask_stats_df.compute())
            zone  mean  max  min   sum       std    var  count
        0     0  22.0   44    0   550  14.21267  202.0     25
        1    10  27.0   49    5   675  14.21267  202.0     25
        2    20  72.0   94   50  1800  14.21267  202.0     25
        3    30  77.0   99   55  1925  14.21267  202.0     25

        >>> # Custom Stats with dask backed xarray DataArrays
        >>> dask_custom_stats ={'double_sum': lambda val: val.sum()*2}
        >>> dask_custom_stats_df = stats(
        >>>      zones=dask_zones, values=dask_values, stats_funcs=custom_stats
        >>> )
        >>> print(type(dask_custom_stats_df))
        <class 'dask.dataframe.core.DataFrame'>
        >>> print(dask_custom_stats_df.compute())
            zone  double_sum
        0     0        1100
        1    10        1350
        2    20        3600
        3    30        3850
    """

    if zones.shape != values.shape:
        raise ValueError("`zones` and `values` must have same shape.")

    if not (
        issubclass(zones.data.dtype.type, np.integer)
        or issubclass(zones.data.dtype.type, np.floating)
    ):
        raise ValueError("`zones` must be an array of integers.")

    if not (
        issubclass(values.data.dtype.type, np.integer)
        or issubclass(values.data.dtype.type, np.floating)
    ):
        raise ValueError("`values` must be an array of integers or floats.")

    if isinstance(stats_funcs, list):
        # create a dict of stats
        stats_funcs_dict = {}

        for stats in stats_funcs:
            func = _DEFAULT_STATS.get(stats, None)
            if func is None:
                err_str = f"Invalid stat name. {stats} option not supported."
                raise ValueError(err_str)

            stats_funcs_dict[stats] = func

    elif isinstance(stats_funcs, dict):
        stats_funcs_dict = stats_funcs.copy()

    if isinstance(values.data, np.ndarray):
        # numpy case
        stats_df = _stats_numpy(
            zones,
            values,
            zone_ids,
            stats_funcs_dict,
            nodata_zones,
            nodata_values
        )
    else:
        # dask case
        stats_df = _stats_dask(
            zones,
            values,
            zone_ids,
            stats_funcs_dict,
            nodata_zones,
            nodata_values
        )

    return stats_df


def _crosstab_dict(zones, values, unique_zones, cats, nodata_values, agg):

    crosstab_dict = {}

    unique_zones = list(map(_to_int, unique_zones))
    crosstab_dict["zone"] = unique_zones

    for i in cats:
        crosstab_dict[i] = []

    for cat_id, cat in enumerate(cats):
        for zone_id in unique_zones:
            # get category cat values in the selected zone
            zone_cat_data = _zone_cat_data(
                zones, values, zone_id, nodata_values, cat, cat_id
            )
            zone_cat_count = _stats_count(zone_cat_data)
            crosstab_dict[cat].append(zone_cat_count)

    if agg == "percentage":
        zone_counts = _stats(
            zones,
            values,
            unique_zones,
            {"count": _stats_count},
            nodata_values
        )["count"]
        for c, cat in enumerate(cats):
            for z in range(len(unique_zones)):
                crosstab_dict[cat][z] = (
                    crosstab_dict[cat][z] / zone_counts[z] * 100
                )  # noqa

    return crosstab_dict


def _crosstab_numpy(
    zones,
    values,
    zone_ids,
    cat_ids,
    nodata_zones,
    nodata_values,
    agg
):

    if cat_ids is not None:
        cats = np.array(cat_ids)
    else:
        # no categories provided, find all possible cats in values raster
        if len(values.shape) == 3:
            # 3D case
            cats = values.indexes[values.dims[0]].values
        else:
            # 2D case
            # mask out all invalid values such as: nan, inf
            cats = da.unique(values.data[da.isfinite(values.data)]).compute()
            cats = sorted(list(set(cats) - set([nodata_values])))

    if zone_ids is None:
        # do not consider zone with nodata values
        unique_zones = np.unique(zones.data[np.isfinite(zones.data)])
        unique_zones = sorted(list(set(unique_zones) - set([nodata_zones])))
    else:
        unique_zones = np.array(zone_ids)

    crosstab_dict = _crosstab_dict(
        zones, values, unique_zones, cats, nodata_values, agg
    )

    crosstab_df = pd.DataFrame(crosstab_dict)

    # name columns
    crosstab_df.columns = crosstab_dict.keys()

    return crosstab_df


def _crosstab_dask(
    zones,
    values,
    zone_ids,
    cat_ids,
    nodata_zones,
    nodata_values,
    agg
):

    if cat_ids is not None:
        cats = np.array(cat_ids)
    else:
        # no categories provided, find all possible cats in values raster
        if len(values.shape) == 3:
            # 3D case
            cats = values.indexes[values.dims[0]].values
        else:
            # 2D case
            # precompute categories
            cats = da.unique(values.data[da.isfinite(values.data)]).compute()
            cats = sorted(list(set(cats) - set([nodata_values])))

    if zone_ids is None:
        # precompute unique zones
        unique_zones = da.unique(zones.data[da.isfinite(zones.data)]).compute()
        # do not consider zone with nodata values
        unique_zones = sorted(list(set(unique_zones) - set([nodata_zones])))
    else:
        unique_zones = np.array(zone_ids)

    crosstab_dict = _crosstab_dict(
        zones, values, unique_zones, cats, nodata_values, agg
    )
    crosstab_dict = {
        stats: da.stack(zonal_stats, axis=0)
        for stats, zonal_stats in crosstab_dict.items()
    }

    # generate dask dataframe
    crosstab_df = dd.concat(
        [dd.from_dask_array(stats) for stats in crosstab_dict.values()], axis=1
    )

    # name columns
    crosstab_df.columns = crosstab_dict.keys()

    return crosstab_df


def crosstab(
    zones: xr.DataArray,
    values: xr.DataArray,
    zone_ids: List[Union[int, float]] = None,
    cat_ids: List[Union[int, float]] = None,
    layer: Optional[int] = None,
    agg: Optional[str] = "count",
    nodata_zones: Optional[Union[int, float]] = None,
    nodata_values: Optional[Union[int, float]] = None,
) -> Union[pd.DataFrame, dd.DataFrame]:
    """
    Calculate cross-tabulated (categorical stats) areas
    between two datasets: a zone dataset `zones`, a value dataset `values`
    (a value raster). Infinite and NaN values in `zones` and `values` will
    be ignored.

    Outputs a pandas DataFrame if `zones` and `values` are numpy backed.
    Outputs a dask DataFrame if `zones` and `values` are dask with
    numpy-backed xarray DataArrays.

    Requires a DataArray with a single data dimension, here called the
    "values", indexed using either 2D or 3D coordinates.

    DataArrays with 3D coordinates are expected to contain values
    distributed over different categories that are indexed by the
    additional coordinate.  Such an array would reduce to the
    2D-coordinate case if collapsed across the categories (e.g. if one
    did ``aggc.sum(dim='cat')`` for a categorical dimension ``cat``).

    Parameters
    ----------
    zones : xr.DataArray
        2D data array of integers or floats.
        A zone is all the cells in a raster that have the same value,
        whether or not they are contiguous. The input `zones` raster defines
        the shape, values, and locations of the zones. An unique field
        in the zone input is specified to define the zones.

    values : xr.DataArray
        2D or 3D data array of integers or floats.
        The input value raster contains the input values used in
        calculating the categorical statistic for each zone.

    zone_ids: List of ints, or floats
        List of zones to be included in calculation. If no zone_ids provided,
        all zones will be used.

    cat_ids: List of ints, or floats
        List of categories to be included in calculation.
        If no cat_ids provided, all categories will be used.

    layer: int, default=0
        index of the categorical dimension layer inside the `values` DataArray.

    agg: str, default = 'count'
        Aggregation method.
        If the data is 2D, available options are: percentage, count.
        If the data is 3D, available option is: count.

    nodata_zones: int, float, default=None
        Nodata value in `zones` raster.
        Cells with `nodata` do not belong to any zone,
        and thus excluded from calculation.

    nodata_values: int, float, default=None
        Nodata value in `values` raster.
        Cells with `nodata` do not belong to any zone,
        and thus excluded from calculation.

    Returns
    -------
    crosstab_df : Union[pandas.DataFrame, dask.dataframe.DataFrame]
        A pandas DataFrame, or an uncomputed dask DataFrame,
        where each column is a categorical value and each row is a zone
        with zone id. Each entry presents the statistics, which computed
        using the specified aggregation method, of the category over the zone.

    Examples
    --------
    .. plot::
       :include-source:

        import dask.array as da
        import numpy as np
        import xarray as xr
        from xrspatial.zonal import crosstab

        values_data = np.asarray([[0, 0, 10, 20],
                                  [0, 0, 0, 10],
                                  [0, np.nan, 20, 50],
                                  [10, 30, 40, np.inf],
                                  [10, 10, 50, 0]])
        values = xr.DataArray(values_data)

        zones_data = np.asarray([[1, 1, 6, 6],
                                 [1, np.nan, 6, 6],
                                 [3, 5, 6, 6],
                                 [3, 5, 7, np.nan],
                                 [3, 7, 7, 0]])
        zones = xr.DataArray(zones_data)

        values_dask = xr.DataArray(da.from_array(values, chunks=(3, 3)))
        zones_dask = xr.DataArray(da.from_array(zones, chunks=(3, 3)))

    .. sourcecode:: python

        >>> # Calculate Crosstab, numpy case
        >>> df = crosstab(zones=zones, values=values)
        >>> print(df)
                zone  0.0  10.0  20.0  30.0  40.0  50.0
            0      0    1     0     0     0     0     0
            1      1    3     0     0     0     0     0
            2      3    1     2     0     0     0     0
            3      5    0     0     0     1     0     0
            4      6    1     2     2     0     0     1
            5      7    0     1     0     0     1     1

        >>> # Calculate Crosstab, dask case
        >>> df = crosstab(zones=zones_dask, values=values_dask)
        >>> print(df)
            Dask DataFrame Structure:
            zone	0.0	10.0	20.0	30.0	40.0	50.0
            npartitions=5
            0	float64	int64	int64	int64	int64	int64	int64
            1	...	...	...	...	...	...	...
            ...	...	...	...	...	...	...	...
            4	...	...	...	...	...	...	...
            5	...	...	...	...	...	...	...
            Dask Name: astype, 1186 tasks
        >>> print(dask_df.compute)
                zone  0.0  10.0  20.0  30.0  40.0  50.0
            0      0    1     0     0     0     0     0
            1      1    3     0     0     0     0     0
            2      3    1     2     0     0     0     0
            3      5    0     0     0     1     0     0
            4      6    1     2     2     0     0     1
            5      7    0     1     0     0     1     1
    """
    if not isinstance(zones, xr.DataArray):
        raise TypeError("zones must be instance of DataArray")

    if not isinstance(values, xr.DataArray):
        raise TypeError("values must be instance of DataArray")

    if zones.ndim != 2:
        raise ValueError("zones must be 2D")

    if not (
        issubclass(zones.data.dtype.type, np.integer)
        or issubclass(zones.data.dtype.type, np.floating)
    ):
        raise ValueError("`zones` must be an xarray of integers or floats")

    if not issubclass(values.data.dtype.type, np.integer) and not issubclass(
        values.data.dtype.type, np.floating
    ):
        raise ValueError("`values` must be an xarray of integers or floats")

    if values.ndim not in [2, 3]:
        raise ValueError("`values` must use either 2D or 3D coordinates.")

    agg_2d = ["percentage", "count"]
    agg_3d = ["count"]

    if values.ndim == 2 and agg not in agg_2d:
        raise ValueError(
            f"`agg` method for 2D data array must be one of following {agg_2d}"
        )

    if values.ndim == 3 and agg not in agg_3d:
        raise ValueError(
            f"`agg` method for 3D data array must be one of following {agg_3d}"
        )

    if len(values.shape) == 3:
        # 3D case
        if layer is None:
            layer = 0
        try:
            values.indexes[values.dims[layer]].values
        except (IndexError, KeyError):
            raise ValueError("Invalid `layer`")

        dims = values.dims
        reshape_dims = [dims[layer]] + [d for d in dims if d != dims[layer]]
        # transpose by that category dimension
        values = values.transpose(*reshape_dims)

        if zones.shape != values.shape[1:]:
            raise ValueError("Incompatible shapes")

    if isinstance(values.data, np.ndarray):
        # numpy case
        crosstab_df = _crosstab_numpy(
            zones, values, zone_ids, cat_ids, nodata_zones, nodata_values, agg
        )
    else:
        # dask case
        crosstab_df = _crosstab_dask(
            zones, values, zone_ids, cat_ids, nodata_zones, nodata_values, agg
        )

    return crosstab_df


def apply(
    zones: xr.DataArray,
    values: xr.DataArray,
    func: Callable,
    nodata: Optional[int] = 0
):
    """
    Apply a function to the `values` agg within zones in `zones` agg.
    Change the agg content.

    Parameters
    ----------
    zones : xr.DataArray
        zones.values is a 2d array of integers. A zone is all the cells
        in a raster that have the same value, whether or not they are
        contiguous. The input zone layer defines the shape, values, and
        locations of the zones. An integer field in the zone input is
        specified to define the zones.

    agg : xr.DataArray
        agg.values is either a 2D or 3D array of integers or floats.
        The input value raster.

    func : callable function to apply.

    nodata: int, default=None
        Nodata value in `zones` raster.
        Cells with `nodata` does not belong to any zone,
        and thus excluded from calculation.

    Examples
    --------
    .. sourcecode:: python

        >>> zones_val = np.array([[1, 1, 0, 2],
        >>>                       [0, 2, 1, 2]])
        >>> zones = xarray.DataArray(zones_val)
        >>> values_val = np.array([[2, -1, 5, 3],
        >>>                        [3, np.nan, 20, 10]])
        >>> agg = xarray.DataArray(values_val)
        >>> func = lambda x: 0
        >>> apply(zones, agg, func)
        >>> agg
        array([[0, 0, 5, 0],
               [3, 0, 0, 0]])
    """
    if not isinstance(zones, xr.DataArray):
        raise TypeError("zones must be instance of DataArray")

    if not isinstance(values, xr.DataArray):
        raise TypeError("values must be instance of DataArray")

    if zones.ndim != 2:
        raise ValueError("zones must be 2D")

    if values.ndim != 2 and values.ndim != 3:
        raise ValueError("values must be either 2D or 3D coordinates")

    if zones.shape != values.shape[:2]:
        raise ValueError("Incompatible shapes between `zones` and `values`")

    if not issubclass(zones.values.dtype.type, np.integer):
        raise ValueError("`zones.values` must be an array of integers")

    if not (
        issubclass(values.values.dtype.type, np.integer)
        or issubclass(values.values.dtype.type, np.floating)
    ):
        raise ValueError("`values` must be an array of integers or float")

    # entries of nodata remain the same
    remain_entries = zones.data == nodata

    # entries with to be included in calculation
    zones_entries = zones.data != nodata

    if len(values.shape) == 3:
        z = values.shape[-1]
        # add new z-dimension in case 3D `values` aggregate
        remain_entries = np.repeat(
            remain_entries[:, :, np.newaxis],
            z,
            axis=-1
        )
        zones_entries = np.repeat(
            zones_entries[:, :, np.newaxis],
            z,
            axis=-1
        )

    remain_mask = np.ma.masked_array(values.data, mask=remain_entries)
    zones_mask = np.ma.masked_array(values.data, mask=zones_entries)

    # apply func to corresponding `values` of `zones`
    vfunc = np.vectorize(func)
    values_func = vfunc(zones_mask)
    values.values = (
        remain_mask.data
        * remain_mask.mask
        + values_func.data
        * values_func.mask
    )


def get_full_extent(crs):
    """
    Returns the full extent of a map projection, available projections
    are 'Mercator' and 'Geographic'.

    Parameters
    ----------
    crs : str
        Input projection.

    Returns
    -------
    extent : tuple
        Projection extent of form ((min_x, max_x), (min_y, max_y)).

    Examples
    --------
    .. sourcecode:: python

        >>> from xrspatial.zonal import get_full_extent

        >>> extent = get_full_extent('Mercator')
        >>> print(extent)
        ((-20000000.0, 20000000.0), (-20000000.0, 20000000.0))
    """
    Mercator = (-20e6, 20e6), (-20e6, 20e6)
    Geographic = (-180, 180), (-90, 90)

    def _crs_code_mapping():
        CRS_CODES = {}
        CRS_CODES["Mercator"] = Mercator
        CRS_CODES["Geographic"] = Geographic
        return CRS_CODES

    CRS_CODES = _crs_code_mapping()
    return CRS_CODES[crs]


def suggest_zonal_canvas(
    smallest_area: Union[int, float],
    x_range: Union[tuple, list],
    y_range: Union[tuple, list],
    crs: str = "Mercator",
    min_pixels: int = 25,
) -> tuple:
    """
    Given a coordinate reference system (crs), a set of polygons with
    corresponding x range and y range, calculate the height and width
    of canvas so that the smallest polygon (polygon with smallest area)
    is rasterized with at least min pixels.

    Currently, we assume that the smallest polygon does not intersect
    others. One should note that a polygon can have different shapes
    when it is rasterized in canvases of different size. Thus, we cannot
    100% guarantee the actual number of pixels after rasterization.
    It is recommended to add an additional of 5% to @min_pixels parameter.

    Parameters
    ----------
    x_range : tuple or list of float or int
        The full x extent of the polygon GeoDataFrame.
    y_range : tuple or list of float or int
        The full y extent of the polygon GeoDataFrame.
    smallest_area : float or int
        Area of the smallest polygon.
    crs : str, default='Mercator'
        Name of the coordinate reference system.
    min_pixels : int, default=25
        Expected number of pixels of the polygon with smallest area
        when the whole dataframe is rasterized.

    Returns
    -------
    height, width: int
        Height and width of the canvas in pixel space.

    Examples
    --------
    .. sourcecode:: python

        >>> # Imports
        >>> from spatialpandas import GeoDataFrame
        >>> import geopandas as gpd
        >>> import datashader as ds

        >>> df = gpd.read_file(gpd.datasets.get_path('naturalearth_lowres'))
        >>> df = df.to_crs("EPSG:3857")
        >>> df = df[df.continent != 'Antarctica']
        >>> df['id'] = [i for i in range(len(df.index))]
        >>> xmin, ymin, xmax, ymax = (
                df.bounds.minx.min(),
                df.bounds.miny.min(),
                df.bounds.maxx.max(),
                df.bounds.maxy.max(),
            )
        >>> x_range = (xmin, xmax)
        >>> y_range = (ymin, ymax)
        >>> smallest_area = df.area.min()
        >>> min_pixels = 20
        >>> height, width = suggest_zonal_canvas(
                x_range=x_range,
                y_range=y_range,
                smallest_area=smallest_area,
                crs='Mercator',
                min_pixels=min_pixels,
            )
        >>> cvs = ds.Canvas(x_range=x_range, y_range=y_range,
        >>>             plot_height=height, plot_width=width)
        >>> spatial_df = GeoDataFrame(df, geometry='geometry')
        >>> agg = cvs.polygons(spatial_df, 'geometry', agg=ds.max('id'))
        >>> min_poly_id = df.area.argmin()
        >>> actual_min_pixels = len(np.where(agg.data==min_poly_id)[0])
    """
    full_xrange, full_yrange = get_full_extent(crs)
    xmin, xmax = full_xrange
    ymin, ymax = full_yrange
    aspect_ratio = (xmax - xmin) / (ymax - ymin)

    # area that a pixel stands for
    pixel_area = smallest_area / min_pixels
    # total_area of whole GeoDataFrame
    total_area = (xmax - xmin) * (ymax - ymin)
    # total pixels needed
    total_pixels = total_area / pixel_area
    # We have, h * w = total_pixels
    # and,     w / h = aspect_ratio
    # Thus,    aspect_ratio * h**2 = total_pixels
    h = sqrt(total_pixels / aspect_ratio)
    w = aspect_ratio * h
    canvas_h = int(h * (y_range[1] - y_range[0]) / (ymax - ymin))
    canvas_w = int(w * (x_range[1] - x_range[0]) / (xmax - xmin))

    return canvas_h, canvas_w


@ngjit
def _area_connectivity(data, n=4):
    out = np.zeros_like(data)
    rows, cols = data.shape
    uid = 1

    src_window = np.zeros(shape=(n,), dtype=data.dtype)
    area_window = np.zeros(shape=(n,), dtype=data.dtype)

    for y in range(0, rows):
        for x in range(0, cols):

            val = data[y, x]

            if np.isnan(val):
                out[y, x] = val
                continue

            if n == 8:
                src_window[0] = data[max(y - 1, 0), max(x - 1, 0)]
                src_window[1] = data[y, max(x - 1, 0)]
                src_window[2] = data[min(y + 1, rows - 1), max(x - 1, 0)]
                src_window[3] = data[max(y - 1, 0), x]
                src_window[4] = data[min(y + 1, rows - 1), x]
                src_window[5] = data[max(y - 1, 0), min(x + 1, cols - 1)]
                src_window[6] = data[y, min(x + 1, cols - 1)]
                src_window[7] = data[min(y + 1, rows - 1), min(x + 1, cols - 1)] # noqa

                area_window[0] = out[max(y - 1, 0), max(x - 1, 0)]
                area_window[1] = out[y, max(x - 1, 0)]
                area_window[2] = out[min(y + 1, rows - 1), max(x - 1, 0)]
                area_window[3] = out[max(y - 1, 0), x]
                area_window[4] = out[min(y + 1, rows - 1), x]
                area_window[5] = out[max(y - 1, 0), min(x + 1, cols - 1)]
                area_window[6] = out[y, min(x + 1, cols - 1)]
                area_window[7] = out[min(y + 1, rows - 1), min(x + 1, cols - 1)] # noqa

            else:
                src_window[0] = data[y, max(x - 1, 0)]
                src_window[1] = data[max(y - 1, 0), x]
                src_window[2] = data[min(y + 1, rows - 1), x]
                src_window[3] = data[y, min(x + 1, cols - 1)]

                area_window[0] = out[y, max(x - 1, 0)]
                area_window[1] = out[max(y - 1, 0), x]
                area_window[2] = out[min(y + 1, rows - 1), x]
                area_window[3] = out[y, min(x + 1, cols - 1)]

            # check in has matching value in neighborhood
            rtol = 1e-05
            atol = 1e-08
            is_close = np.abs(src_window - val) <= (atol + rtol * np.abs(val))
            neighbor_matches = np.where(is_close)[0]

            if len(neighbor_matches) > 0:

                # check in has area already assigned
                assigned_value = None
                for j in range(len(neighbor_matches)):
                    area_val = area_window[neighbor_matches[j]]
                    if area_val > 0:
                        assigned_value = area_val
                        break

                if assigned_value is not None:
                    out[y, x] = assigned_value
                else:
                    out[y, x] = uid
                    uid += 1
            else:
                out[y, x] = uid
                uid += 1

    for y in range(0, rows):
        for x in range(0, cols):

            if n == 8:
                src_window[0] = data[max(y - 1, 0), max(x - 1, 0)]
                src_window[1] = data[y, max(x - 1, 0)]
                src_window[2] = data[min(y + 1, rows - 1), max(x - 1, 0)]
                src_window[3] = data[max(y - 1, 0), x]
                src_window[4] = data[min(y + 1, rows - 1), x]
                src_window[5] = data[max(y - 1, 0), min(x + 1, cols - 1)]
                src_window[6] = data[y, min(x + 1, cols - 1)]
                src_window[7] = data[min(y + 1, rows - 1), min(x + 1, cols - 1)] # noqa

                area_window[0] = out[max(y - 1, 0), max(x - 1, 0)]
                area_window[1] = out[y, max(x - 1, 0)]
                area_window[2] = out[min(y + 1, rows - 1), max(x - 1, 0)]
                area_window[3] = out[max(y - 1, 0), x]
                area_window[4] = out[min(y + 1, rows - 1), x]
                area_window[5] = out[max(y - 1, 0), min(x + 1, cols - 1)]
                area_window[6] = out[y, min(x + 1, cols - 1)]
                area_window[7] = out[min(y + 1, rows - 1), min(x + 1, cols - 1)] # noqa

            else:
                src_window[0] = data[y, max(x - 1, 0)]
                src_window[1] = data[max(y - 1, 0), x]
                src_window[2] = data[min(y + 1, rows - 1), x]
                src_window[3] = data[y, min(x + 1, cols - 1)]

                area_window[0] = out[y, max(x - 1, 0)]
                area_window[1] = out[max(y - 1, 0), x]
                area_window[2] = out[min(y + 1, rows - 1), x]
                area_window[3] = out[y, min(x + 1, cols - 1)]

            val = data[y, x]

            if np.isnan(val):
                continue

            # check in has matching value in neighborhood
            rtol = 1e-05
            atol = 1e-08
            is_close = np.abs(src_window - val) <= (atol + rtol * np.abs(val))
            neighbor_matches = np.where(is_close)[0]

            # check in has area already assigned
            assigned_values_min = None
            for j in range(len(neighbor_matches)):
                area_val = area_window[neighbor_matches[j]]
                nn = assigned_values_min is not None
                if nn and assigned_values_min != area_val:
                    if assigned_values_min > area_val:

                        # replace
                        for y1 in range(0, rows):
                            for x1 in range(0, cols):
                                if out[y1, x1] == assigned_values_min:
                                    out[y1, x1] = area_val

                        assigned_values_min = area_val

                    else:
                        # replace
                        for y1 in range(0, rows):
                            for x1 in range(0, cols):
                                if out[y1, x1] == area_val:
                                    out[y1, x1] = assigned_values_min

                elif assigned_values_min is None:
                    assigned_values_min = area_val

    return out


def regions(
    raster: xr.DataArray, neighborhood: int = 4, name: str = "regions"
) -> xr.DataArray:
    """
    Create unique regions of raster based on pixel value connectivity.
    Connectivity can be based on either 4 or 8-pixel neighborhoods.
    Output raster contain a unique int for each connected region.

    Parameters
    ----------
    raster : xr.DataArray
    connections : int, default=4
        4 or 8 pixel-based connectivity.
    name: str, default='regions'
        output xr.DataArray.name property.

    Returns
    -------
    regions_agg : xarray.DataArray

    References
    ----------
        - Tomislav Hengl: http://spatial-analyst.net/ILWIS/htm/ilwisapp/areanumbering_algorithm.htm # noqa

    Examples
    --------
    .. plot::
       :include-source:

        import datashader as ds
        import matplotlib.pyplot as plt
        from xrspatial import generate_terrain
        from xrspatial.zonal import regions

        # Create Canvas
        W = 500
        H = 300
        cvs = ds.Canvas(plot_width = W,
                        plot_height = H,
                        x_range = (-20e6, 20e6),
                        y_range = (-20e6, 20e6))


        # Generate Values
        terrain_agg = generate_terrain(canvas = cvs)

        # Edit Attributes
        terrain_agg = terrain_agg.assign_attrs(
            {
                'Description': 'Example Terrain',
                'units': 'km',
                'Max Elevation': '4000',
            }
        )

        terrain_agg = terrain_agg.rename({'x': 'lon', 'y': 'lat'})
        terrain_agg = terrain_agg.rename('Elevation')

        # Create Regions
        regions_agg = regions(terrain_agg)

        # Edit Attributes
        regions_agg = regions_agg.assign_attrs({'Description': 'Example Regions',
                                                 'units': ''})
        regions_agg = regions_agg.rename('Region Value')

        # Plot Terrain (Values)
        terrain_agg.plot(cmap = 'terrain', aspect = 2, size = 4)
        plt.title("Terrain (Values)")
        plt.ylabel("latitude")
        plt.xlabel("longitude")

        # Plot Regions
        regions_agg.plot(cmap = 'terrain', aspect = 2, size = 4)
        plt.title("Regions")
        plt.ylabel("latitude")
        plt.xlabel("longitude")

    .. sourcecode:: python

        >>> print(terrain_agg[200:203, 200:202])
        <xarray.DataArray 'Elevation' (lat: 3, lon: 2)>
        array([[1264.02249454, 1261.94748873],
               [1285.37061171, 1282.48046696],
               [1306.02305679, 1303.40657515]])
        Coordinates:
          * lon      (lon) float64 -3.96e+06 -3.88e+06
          * lat      (lat) float64 6.733e+06 6.867e+06 7e+06
        Attributes:
            res:            1
            Description:    Example Terrain
            units:          km
            Max Elevation:  4000

        >>> print(regions_agg[200:203, 200:202])
        <xarray.DataArray 'Region Value' (lat: 3, lon: 2)>
        array([[39557., 39558.],
               [39943., 39944.],
               [40327., 40328.]])
        Coordinates:
          * lon      (lon) float64 -3.96e+06 -3.88e+06
          * lat      (lat) float64 6.733e+06 6.867e+06 7e+06
        Attributes:
            res:            1
            Description:    Example Regions
            units:
            Max Elevation:  4000
    """
    if neighborhood not in (4, 8):
        raise ValueError("`neighborhood` value must be either 4 or 8)")

    out = _area_connectivity(raster.data, n=neighborhood)

    return DataArray(
        out,
        name=name,
        dims=raster.dims,
        coords=raster.coords,
        attrs=raster.attrs
    )


def _bool_crop(arr, rows_flags, cols_flags):
    top = np.argwhere(rows_flags).flatten()[0]
    bottom = np.argwhere(rows_flags).flatten()[-1]
    left = np.argwhere(cols_flags).flatten()[0]
    right = np.argwhere(cols_flags).flatten()[-1]
    return arr[top: bottom + 1, left: right + 1]


@ngjit
def _trim(data, excludes):

    rows, cols = data.shape

    # find empty top rows
    top = 0
    scan_complete = False
    for y in range(rows):

        if scan_complete:
            break

        top = y
        for x in range(cols):
            val = data[y, x]
            is_nodata = False
            for e in excludes:
                if e == val:
                    is_nodata = True
                    break

            if not is_nodata:
                scan_complete = True
                break

    # find empty bottom rows
    bottom = 0
    scan_complete = False
    for y in range(rows - 1, -1, -1):
        if scan_complete:
            break
        bottom = y
        for x in range(cols):
            val = data[y, x]
            is_nodata = False
            for e in excludes:
                if e == val:
                    is_nodata = True
                    break
            if not is_nodata:
                scan_complete = True
                break

    # find empty left cols
    left = 0
    scan_complete = False
    for x in range(cols):
        if scan_complete:
            break
        left = x
        for y in range(rows):
            val = data[y, x]
            is_nodata = False
            for e in excludes:
                if e == val:
                    is_nodata = True
                    break
            if not is_nodata:
                scan_complete = True
                break

    # find empty right cols
    right = 0
    scan_complete = False
    for x in range(cols - 1, -1, -1):
        if scan_complete:
            break
        right = x
        for y in range(rows):
            val = data[y, x]
            is_nodata = False
            for e in excludes:
                if e == val:
                    is_nodata = True
                    break
            if not is_nodata:
                scan_complete = True
                break

    return top, bottom, left, right


def trim(
    raster: xr.DataArray,
    values: Union[list, tuple] = (np.nan,),
    name: str = "trim"
) -> xr.DataArray:
    """
    Trim scans from the edges and eliminates rows / cols which only
    contain the values supplied.

    Parameters
    ----------
    raster: xr.DataArray
    values: list or tuple, default=(np.nan)
        List of zone ids to trim from raster edge.
    name: str, default='trim'
        Output xr.DataArray.name property.

    Returns
    -------
    trim_agg : xarray.DataArray

    Notes
    -----
        - This operation will change the output size of the raster.

    Examples
    --------
    .. plot::
       :include-source:

        import datashader as ds
        import numpy as np
        import matplotlib.pyplot as plt
        from xrspatial import generate_terrain
        from xrspatial.zonal import trim

        # Create Canvas
        W = 500
        H = 300
        cvs = ds.Canvas(plot_width = W,
                        plot_height = H,
                        x_range = (-20e6, 20e6),
                        y_range = (-20e6, 20e6))


        # Generate Terrain
        terrain_agg = generate_terrain(canvas = cvs)

        # Edit Attributes
        terrain_agg = terrain_agg.assign_attrs(
            {
                'Description': 'Example Terrain',
                'units': 'km',
                'Max Elevation': '4000',
            }
        )

        terrain_agg = terrain_agg.rename({'x': 'lon', 'y': 'lat'})
        terrain_agg = terrain_agg.rename('Elevation')
        terrain_agg = terrain_agg.astype('int')

        # Trim Image
        trimmed_agg = trim(raster = terrain_agg, values = [0])

        # Edit Attributes
        trimmed_agg = trimmed_agg.assign_attrs({'Description': 'Example Trim'})

        # Plot Terrain
        terrain_agg.plot(cmap = 'terrain', aspect = 2, size = 4)
        plt.title("Terrain")
        plt.ylabel("latitude")
        plt.xlabel("longitude")

        # Plot Trimmed Terrain
        trimmed_agg.plot(cmap = 'terrain', aspect = 2, size = 4)
        plt.title("Trim")
        plt.ylabel("latitude")
        plt.xlabel("longitude")

    .. sourcecode:: python

        >>> print(terrain_agg.shape)
        (300, 500)

        >>> print(terrain_agg.attrs)
        {
            'res': 1,
            'Description': 'Example Terrain',
            'units': 'km',
            'Max Elevation': '4000',
        }

        >>> print(trimmed_agg.shape)
        (268, 500)

        >>> print(trimmed_agg.attrs)
        {
            'res': 1,
            'Description': 'Example Trim',
            'units': 'km',
            'Max Elevation': '4000',
        }
    """
    top, bottom, left, right = _trim(raster.data, values)
    arr = raster[top: bottom + 1, left: right + 1]
    arr.name = name
    return arr


@ngjit
def _crop(data, values):

    rows, cols = data.shape

    top = -1
    bottom = -1
    left = -1
    right = -1

    # find empty top rows
    top = 0
    scan_complete = False
    for y in range(rows):

        if scan_complete:
            break

        top = y

        for x in range(cols):
            val = data[y, x]
            for v in values:
                if v == val:
                    scan_complete = True
                    break
                else:
                    continue

            if scan_complete:
                break

    # find empty bottom rows
    bottom = 0
    scan_complete = False
    for y in range(rows - 1, -1, -1):

        if scan_complete:
            break

        bottom = y

        for x in range(cols):
            val = data[y, x]
            for e in values:
                if e == val:
                    scan_complete = True
                    break
                else:
                    continue

            if scan_complete:
                break

    # find empty left cols
    left = 0
    scan_complete = False
    for x in range(cols):

        if scan_complete:
            break

        left = x

        for y in range(rows):
            val = data[y, x]
            for e in values:
                if e == val:
                    scan_complete = True
                    break
                else:
                    continue

            if scan_complete:
                break

    # find empty right cols
    right = 0
    scan_complete = False
    for x in range(cols - 1, -1, -1):
        if scan_complete:
            break
        right = x
        for y in range(rows):
            val = data[y, x]
            for e in values:
                if e == val:
                    scan_complete = True
                    break
                else:
                    continue

            if scan_complete:
                break

    return top, bottom, left, right


def crop(
    zones: xr.DataArray,
    values: xr.DataArray,
    zones_ids: Union[list, tuple],
    name: str = "crop",
):
    """
    Crop scans from edges and eliminates rows / cols until one of the
    input values is found.

    Parameters
    ----------
    zones : xr.DataArray
        Input zone raster.

    values: xr.DataArray
        Input values raster.

    zones_ids : list or tuple
        List of zone ids to crop raster.

    name: str, default='crop'
        Output xr.DataArray.name property.

    Returns
    -------
    crop_agg : xarray.DataArray

    Notes
    -----
        - This operation will change the output size of the raster.

    Examples
    --------
    .. plot::
       :include-source:

        import datashader as ds
        import matplotlib.pyplot as plt
        from xrspatial import generate_terrain
        from xrspatial.zonal import crop

        # Create Canvas
        W = 500
        H = 300
        cvs = ds.Canvas(plot_width = W,
                        plot_height = H,
                        x_range = (-20e6, 20e6),
                        y_range = (-20e6, 20e6))


        # Generate Zones
        terrain_agg = generate_terrain(canvas = cvs)

        # Edit Attributes
        terrain_agg = terrain_agg.assign_attrs(
            {
                'Description': 'Example Terrain',
                'units': 'km',
                'Max Elevation': '4000',
            }
        )

        terrain_agg = terrain_agg.rename({'x': 'lon', 'y': 'lat'})
        terrain_agg = terrain_agg.rename('Elevation')

        # Crop Image
        values_agg = terrain_agg[0:300, 0:250]
        cropped_agg = crop(
            zones=terrain_agg,
            values=values_agg,
            zones_ids=[0],
        )

        # Edit Attributes
        cropped_agg = cropped_agg.assign_attrs({'Description': 'Example Crop'})

        # Plot Terrain
        terrain_agg.plot(cmap = 'terrain', aspect = 2, size = 4)
        plt.title("Terrain")
        plt.ylabel("latitude")
        plt.xlabel("longitude")

        # Plot Cropped Terrain
        cropped_agg.plot(cmap = 'terrain', aspect = 2, size = 4)
        plt.title("Crop")
        plt.ylabel("latitude")
        plt.xlabel("longitude")

    .. sourcecode:: python

        >>> print(terrain_agg.shape)
        (300, 500)

        >>> print(terrain_agg.attrs)
        {
            'res': 1,
            'Description': 'Example Terrain',
            'units': 'km',
            'Max Elevation': '4000',
        }

        >>> print(cropped_agg.shape)
        (300, 250)

        >>> print(cropped_agg.attrs)
        {
            'res': 1,
            'Description': 'Example Crop',
            'units': 'km',
            'Max Elevation': '4000',
        }
    """
    top, bottom, left, right = _crop(zones.data, zones_ids)
    arr = values[top: bottom + 1, left: right + 1]
    arr.name = name
    return arr



import pytest
import xarray as xr
import numpy as np

import dask.array as da

from xrspatial.utils import doesnt_have_cuda
from xrspatial.utils import ngjit

from xrspatial import mean
from xrspatial.focal import hotspots, apply, focal_stats
from xrspatial.convolution import (
    convolve_2d, calc_cellsize, circle_kernel, annulus_kernel
)


def _do_sparse_array(data_array):
    import random
    indx = list(zip(*np.where(data_array)))
    pos = random.sample(range(data_array.size), data_array.size//2)
    indx = np.asarray(indx)[pos]
    r = indx[:, 0]
    c = indx[:, 1]
    data_half = data_array.copy()
    data_half[r, c] = 0
    return data_half


def _do_gaussian_array():
    _x = np.linspace(0, 50, 101)
    _y = _x.copy()
    _mean = 25
    _sdev = 5
    X, Y = np.meshgrid(_x, _y, sparse=True)
    x_fac = -np.power(X-_mean, 2)
    y_fac = -np.power(Y-_mean, 2)
    gaussian = np.exp((x_fac+y_fac)/(2*_sdev**2)) / (2.5*_sdev)
    return gaussian


data_random = np.random.random_sample((100, 100))
data_random_sparse = _do_sparse_array(data_random)
data_gaussian = _do_gaussian_array()


def test_mean_transfer_function_cpu():
    # numpy case
    numpy_agg = xr.DataArray(data_random)
    numpy_mean = mean(numpy_agg)
    assert isinstance(numpy_mean.data, np.ndarray)

    # dask + numpy case
    dask_numpy_agg = xr.DataArray(da.from_array(data_random, chunks=(3, 3)))
    dask_numpy_mean = mean(dask_numpy_agg)
    assert isinstance(dask_numpy_mean.data, da.Array)

    # both output same results
    assert np.isclose(
        numpy_mean, dask_numpy_mean.compute(), equal_nan=True
    ).all()
    assert numpy_agg.shape == numpy_mean.shape


@pytest.mark.skipif(doesnt_have_cuda(), reason="CUDA Device not Available")
def test_mean_transfer_function_gpu_equals_cpu():

    import cupy

    # cupy case
    cupy_agg = xr.DataArray(cupy.asarray(data_random))
    cupy_mean = mean(cupy_agg)
    assert isinstance(cupy_mean.data, cupy.ndarray)

    # numpy case
    numpy_agg = xr.DataArray(data_random)
    numpy_mean = mean(numpy_agg)

    assert np.isclose(numpy_mean, cupy_mean.data.get(), equal_nan=True).all()

    # dask + cupy case not implemented
    dask_cupy_agg = xr.DataArray(
        da.from_array(cupy.asarray(data_random), chunks=(3, 3))
    )
    with pytest.raises(NotImplementedError) as e_info:
        mean(dask_cupy_agg)
        assert e_info


convolve_2d_data = np.array([[0., 1., 1., 1., 1., 1.],
                             [1., 0., 1., 1., 1., 1.],
                             [1., 1., 0., 1., 1., 1.],
                             [1., 1., 1., np.nan, 1., 1.],
                             [1., 1., 1., 1., 0., 1.],
                             [1., 1., 1., 1., 1., 0.]])


def test_kernel():
    data = convolve_2d_data
    m, n = data.shape
    agg = xr.DataArray(data, dims=['y', 'x'])
    agg['x'] = np.linspace(0, n, n)
    agg['y'] = np.linspace(0, m, m)

    cellsize_x, cellsize_y = calc_cellsize(agg)

    kernel1 = circle_kernel(cellsize_x, cellsize_y, 2)
    expected_kernel1 = np.array([[0, 1, 0],
                                 [1, 1, 1],
                                 [0, 1, 0]])
    assert isinstance(kernel1, np.ndarray)
    assert np.isclose(kernel1, expected_kernel1, equal_nan=True).all()

    kernel2 = annulus_kernel(cellsize_x, cellsize_y, 2, 0.5)
    expected_kernel2 = np.array([[0, 1, 0],
                                 [1, 0, 1],
                                 [0, 1, 0]])
    assert isinstance(kernel2, np.ndarray)
    assert np.isclose(kernel2, expected_kernel2, equal_nan=True).all()


def test_convolution():
    data = convolve_2d_data
    dask_data = da.from_array(data, chunks=(3, 3))

    kernel1 = np.ones((1, 1))
    numpy_output_1 = convolve_2d(data, kernel1)
    expected_output_1 = np.array([[0., 1., 1., 1., 1., 1.],
                                  [1., 0., 1., 1., 1., 1.],
                                  [1., 1., 0., 1., 1., 1.],
                                  [1., 1., 1., np.nan, 1., 1.],
                                  [1., 1., 1., 1., 0., 1.],
                                  [1., 1., 1., 1., 1., 0.]])
    assert isinstance(numpy_output_1, np.ndarray)
    assert np.isclose(numpy_output_1, expected_output_1, equal_nan=True).all()

    dask_output_1 = convolve_2d(dask_data, kernel1)
    assert isinstance(dask_output_1, da.Array)
    assert np.isclose(
        dask_output_1.compute(), expected_output_1, equal_nan=True
    ).all()

    kernel2 = np.array([[0, 1, 0],
                        [1, 1, 1],
                        [0, 1, 0]])
    numpy_output_2 = convolve_2d(data, kernel2)
    expected_output_2 = np.array([
        [np.nan, np.nan, np.nan, np.nan, np.nan, np.nan],
        [np.nan, 4., 3., 5., 5., np.nan],
        [np.nan, 3., np.nan, np.nan, np.nan, np.nan],
        [np.nan, 5., np.nan, np.nan, np.nan, np.nan],
        [np.nan, 5., np.nan, np.nan, np.nan, np.nan],
        [np.nan, np.nan, np.nan, np.nan, np.nan, np.nan]
    ])
    # kernel2 is of 3x3, thus the border edge is 1 cell long.
    # currently, ignoring border edge (i.e values in edges are all nans)
    assert isinstance(numpy_output_2, np.ndarray)
    assert np.isclose(
        numpy_output_2, expected_output_2, equal_nan=True
    ).all()

    dask_output_2 = convolve_2d(dask_data, kernel2)
    assert isinstance(dask_output_2, da.Array)
    assert np.isclose(
        dask_output_2.compute(), expected_output_2, equal_nan=True
    ).all()

    kernel3 = np.array([[0, 1, 0],
                        [1, 0, 1],
                        [0, 1, 0]])
    numpy_output_3 = convolve_2d(data, kernel3)
    expected_output_3 = np.array([
        [np.nan, np.nan, np.nan, np.nan, np.nan, np.nan],
        [np.nan, 4., 2., 4., 4., np.nan],
        [np.nan, 2., np.nan, np.nan, np.nan, np.nan],
        [np.nan, 4., np.nan, np.nan, np.nan, np.nan],
        [np.nan, 4., np.nan, np.nan, np.nan, np.nan],
        [np.nan, np.nan, np.nan, np.nan, np.nan, np.nan]
    ])
    # kernel3 is of 3x3, thus the border edge is 1 cell long.
    # currently, ignoring border edge (i.e values in edges are all nans)
    assert isinstance(numpy_output_3, np.ndarray)
    assert np.isclose(numpy_output_3, expected_output_3, equal_nan=True).all()

    dask_output_3 = convolve_2d(dask_data, kernel3)
    assert isinstance(dask_output_3, da.Array)
    assert np.isclose(
        dask_output_3.compute(), expected_output_3, equal_nan=True
    ).all()


@pytest.mark.skipif(doesnt_have_cuda(), reason="CUDA Device not Available")
def test_2d_convolution_gpu_equals_cpu():

    import cupy

    data = convolve_2d_data
    numpy_agg = xr.DataArray(data)
    cupy_agg = xr.DataArray(cupy.asarray(data))

    kernel1 = np.ones((1, 1))
    output_numpy1 = convolve_2d(numpy_agg.data, kernel1)
    output_cupy1 = convolve_2d(cupy_agg.data, kernel1)
    assert isinstance(output_cupy1, cupy.ndarray)
    assert np.isclose(output_numpy1, output_cupy1.get(), equal_nan=True).all()

    kernel2 = np.array([[0, 1, 0],
                        [1, 1, 1],
                        [0, 1, 0]])
    output_numpy2 = convolve_2d(numpy_agg.data, kernel2)
    output_cupy2 = convolve_2d(cupy_agg.data, kernel2)
    assert isinstance(output_cupy2, cupy.ndarray)
    assert np.isclose(output_numpy2, output_cupy2.get(), equal_nan=True).all()

    kernel3 = np.array([[0, 1, 0],
                        [1, 0, 1],
                        [0, 1, 0]])
    output_numpy3 = convolve_2d(numpy_agg.data, kernel3)
    output_cupy3 = convolve_2d(cupy_agg.data, kernel3)
    assert isinstance(output_cupy3, cupy.ndarray)
    assert np.isclose(output_numpy3, output_cupy3.get(), equal_nan=True).all()

    # dask + cupy case not implemented
    dask_cupy_agg = xr.DataArray(
        da.from_array(cupy.asarray(data), chunks=(3, 3))
    )
    with pytest.raises(NotImplementedError) as e_info:
        convolve_2d(dask_cupy_agg.data, kernel3)
        assert e_info


data_apply = np.array([[0, 1, 2, 3, 4, 5],
                       [6, 7, 8, 9, 10, 11],
                       [12, 13, 14, 15, 16, 17],
                       [18, 19, 20, 21, 22, 23]])

kernel_apply = np.array([[0, 1, 0], [1, 0, 1], [0, 1, 0]])


def test_apply_cpu():
    @ngjit
    def func_zero_cpu(x):
        return 0

    # numpy case
    numpy_agg = xr.DataArray(data_apply)
    numpy_apply = apply(numpy_agg, kernel_apply, func_zero_cpu)
    assert isinstance(numpy_apply.data, np.ndarray)
    assert numpy_agg.shape == numpy_apply.shape
    assert np.count_nonzero(numpy_apply.data) == 0

    # dask + numpy case
    dask_numpy_agg = xr.DataArray(da.from_array(data_apply, chunks=(3, 3)))
    dask_numpy_apply = apply(dask_numpy_agg, kernel_apply, func_zero_cpu)
    assert isinstance(dask_numpy_apply.data, da.Array)

    # both output same results
    assert np.isclose(
        numpy_apply, dask_numpy_apply.compute(), equal_nan=True
    ).all()


@pytest.mark.skipif(doesnt_have_cuda(), reason="CUDA Device not Available")
def test_apply_gpu_equals_gpu():
    def func_zero(x):
        return 0

    @ngjit
    def func_zero_cpu(x):
        return 0

    # cupy case
    import cupy
    cupy_agg = xr.DataArray(cupy.asarray(data_apply))
    cupy_apply = apply(cupy_agg, kernel_apply, func_zero)
    assert isinstance(cupy_apply.data, cupy.ndarray)
    # numpy case
    numpy_agg = xr.DataArray(data_apply)
    numpy_apply = apply(numpy_agg, kernel_apply, func_zero_cpu)
    assert np.isclose(numpy_apply, cupy_apply.data.get(), equal_nan=True).all()

    # dask + cupy case not implemented
    dask_cupy_agg = xr.DataArray(
        da.from_array(cupy.asarray(data_apply), chunks=(3, 3))
    )
    with pytest.raises(NotImplementedError) as e_info:
        apply(dask_cupy_agg, kernel_apply, func_zero)
        assert e_info


def test_focal_stats_cpu():
    data = np.arange(16).reshape(4, 4)
    numpy_agg = xr.DataArray(data)
    dask_numpy_agg = xr.DataArray(da.from_array(data, chunks=(3, 3)))

    cellsize = (1, 1)
    kernel = circle_kernel(*cellsize, 1.5)

    numpy_focalstats = focal_stats(numpy_agg, kernel)
    assert isinstance(numpy_focalstats.data, np.ndarray)
    assert numpy_focalstats.ndim == 3
    assert numpy_agg.shape == numpy_focalstats.shape[1:]

    dask_numpy_focalstats = focal_stats(dask_numpy_agg, kernel)
    assert isinstance(dask_numpy_focalstats.data, da.Array)

    assert np.isclose(
        numpy_focalstats, dask_numpy_focalstats.compute(), equal_nan=True
    ).all()


def test_hotspot():
    n, m = 10, 10
    data = np.zeros((n, m), dtype=float)

    all_idx = zip(*np.where(data == 0))

    nan_cells = [(i, i) for i in range(m)]
    for cell in nan_cells:
        data[cell[0], cell[1]] = np.nan

    # add some extreme values
    hot_region = [(1, 1), (1, 2), (1, 3),
                  (2, 1), (2, 2), (2, 3),
                  (3, 1), (3, 2), (3, 3)]
    cold_region = [(7, 7), (7, 8), (7, 9),
                   (8, 7), (8, 8), (8, 9),
                   (9, 7), (9, 8), (9, 9)]
    for p in hot_region:
        data[p[0], p[1]] = 10000
    for p in cold_region:
        data[p[0], p[1]] = -10000

    numpy_agg = xr.DataArray(data, dims=['y', 'x'])
    numpy_agg['x'] = np.linspace(0, n, n)
    numpy_agg['y'] = np.linspace(0, m, m)
    cellsize_x, cellsize_y = calc_cellsize(numpy_agg)

    kernel = circle_kernel(cellsize_x, cellsize_y, 2.0)

    no_significant_region = [id for id in all_idx if id not in hot_region and
                             id not in cold_region]

    # numpy case
    numpy_hotspots = hotspots(numpy_agg, kernel)

    # dask + numpy
    dask_numpy_agg = xr.DataArray(da.from_array(data, chunks=(3, 3)))
    dask_numpy_hotspots = hotspots(dask_numpy_agg, kernel)

    assert isinstance(dask_numpy_hotspots.data, da.Array)

    # both output same results
    assert np.isclose(numpy_hotspots.data, dask_numpy_hotspots.data.compute(),
                      equal_nan=True).all()

    # check output's properties
    # output must be an xarray DataArray
    assert isinstance(numpy_hotspots, xr.DataArray)
    assert isinstance(numpy_hotspots.values, np.ndarray)
    assert issubclass(numpy_hotspots.values.dtype.type, np.int8)

    # shape, dims, coords, attr preserved
    assert numpy_agg.shape == numpy_hotspots.shape
    assert numpy_agg.dims == numpy_hotspots.dims
    assert numpy_agg.attrs == numpy_hotspots.attrs
    for coord in numpy_agg.coords:
        assert np.all(numpy_agg[coord] == numpy_hotspots[coord])

    # no nan in output
    assert not np.isnan(np.min(numpy_hotspots))

    # output of extreme regions are non-zeros
    # hot spots
    hot_spot = np.asarray([numpy_hotspots[p] for p in hot_region])
    assert np.all(hot_spot >= 0)
    assert np.sum(hot_spot) > 0
    # cold spots
    cold_spot = np.asarray([numpy_hotspots[p] for p in cold_region])
    assert np.all(cold_spot <= 0)
    assert np.sum(cold_spot) < 0
    # output of no significant regions are 0s
    no_sign = np.asarray([numpy_hotspots[p] for p in no_significant_region])
    assert np.all(no_sign == 0)


@pytest.mark.skipif(doesnt_have_cuda(), reason="CUDA Device not Available")
def test_hotspot_gpu_equals_cpu():
    n, m = 10, 10
    data = np.zeros((n, m), dtype=float)

    nan_cells = [(i, i) for i in range(m)]
    for cell in nan_cells:
        data[cell[0], cell[1]] = np.nan

    # add some extreme values
    hot_region = [(1, 1), (1, 2), (1, 3),
                  (2, 1), (2, 2), (2, 3),
                  (3, 1), (3, 2), (3, 3)]
    cold_region = [(7, 7), (7, 8), (7, 9),
                   (8, 7), (8, 8), (8, 9),
                   (9, 7), (9, 8), (9, 9)]
    for p in hot_region:
        data[p[0], p[1]] = 10000
    for p in cold_region:
        data[p[0], p[1]] = -10000

    numpy_agg = xr.DataArray(data, dims=['y', 'x'])
    numpy_agg['x'] = np.linspace(0, n, n)
    numpy_agg['y'] = np.linspace(0, m, m)

    cellsize_x, cellsize_y = calc_cellsize(numpy_agg)
    kernel = circle_kernel(cellsize_x, cellsize_y, 2.0)
    # numpy case
    numpy_hotspots = hotspots(numpy_agg, kernel)

    # cupy case
    import cupy

    cupy_agg = xr.DataArray(cupy.asarray(data))
    cupy_hotspots = hotspots(cupy_agg, kernel)

    assert isinstance(cupy_hotspots.data, cupy.ndarray)
    assert np.isclose(
        numpy_hotspots, cupy_hotspots.data.get(), equal_nan=True
    ).all()

    # dask + cupy case not implemented
    dask_cupy_agg = xr.DataArray(
        da.from_array(cupy.asarray(data), chunks=(3, 3))
    )
    with pytest.raises(NotImplementedError) as e_info:
        hotspots(dask_cupy_agg, kernel)
        assert e_info



import xarray as xa
import pytest

import datashader as ds
from xrspatial import viewshed

import numpy as np
import pandas as pd

# create an empty image of size 5*5
H = 5
W = 5

canvas = ds.Canvas(plot_width=W, plot_height=H,
                   x_range=(-20, 20), y_range=(-20, 20))

empty_df = pd.DataFrame({
   'x': np.random.normal(.5, 1, 0),
   'y': np.random.normal(.5, 1, 0)
})
empty_agg = canvas.points(empty_df, 'x', 'y')

# coordinates
xs = empty_agg.coords['x'].values
ys = empty_agg.coords['x'].values

# define some values for observer's elevation to test
OBS_ELEVS = [-1, 0, 1]

TERRAIN_ELEV_AT_VP = [-1, 0, 1]


def test_viewshed_invalid_x_view():
    OBSERVER_X = xs[0] - 1
    OBSERVER_Y = 0
    with pytest.raises(Exception) as e_info:  # NOQA
        viewshed(raster=empty_agg, x=OBSERVER_X, y=OBSERVER_Y,
                 observer_elev=10)


def test_viewshed_invalid_y_view():
    OBSERVER_X = 0
    OBSERVER_Y = ys[-1] + 1
    with pytest.raises(Exception) as e_info:  # NOQA
        viewshed(raster=empty_agg, x=OBSERVER_X, y=OBSERVER_Y,
                 observer_elev=10)


def test_viewshed_output_properties():
    for obs_elev in OBS_ELEVS:
        OBSERVER_X = xs[0]
        OBSERVER_Y = ys[0]
        v = viewshed(raster=empty_agg, x=OBSERVER_X, y=OBSERVER_Y,
                     observer_elev=obs_elev)

        assert v.shape[0] == empty_agg.shape[0]
        assert v.shape[1] == empty_agg.shape[1]
        assert isinstance(v, xa.DataArray)
        assert isinstance(v.values, np.ndarray)
        assert type(v.values[0, 0]) == np.float64


def test_viewshed():

    # check if a matrix is symmetric
    def check_symmetric(matrix, rtol=1e-05, atol=1e-08):
        return np.allclose(matrix, matrix.T, rtol=rtol, atol=atol)

    def get_matrices(y, x, height, width):
        # indexing 0 1 ... height-1 and 0 1 ... width-1
        height = height - 1
        width = width - 1

        # find first matrix's diagonal
        tmp = min(y, x)
        f_top_y, f_left_x = y - tmp, x - tmp

        tmp = min(height - y, width - x)
        f_bottom_y, f_right_x = y + tmp, x + tmp

        # find second matrix's antidiagonal
        tmp = min(y, width - x)
        s_top_y, s_right_x = y - tmp, x + tmp

        tmp = min(height - y, x)
        s_bottom_y, s_left_x = y + tmp, x - tmp

        return (f_top_y, f_left_x, f_bottom_y + 1, f_right_x + 1), \
               (s_top_y, s_left_x, s_bottom_y + 1, s_right_x + 1)

    # test on 3 scenarios:
    #   empty image.
    #   image with all 0s, except 1 cell with a negative value.
    #   image with all 0s, except 1 cell with a positive value.

    # for each scenario:
    #   if not empty image,
    #      observer is located at the same position as the non zero value.
    #   observer elevation can be: negative, zero, or positive.

    # assertion:
    #   angle at viewpoint is always 180.
    #   when the observer is above the terrain, all cells are visible.
    #   the symmetric property of observer's visibility.

    for obs_elev in OBS_ELEVS:
        for elev_at_vp in TERRAIN_ELEV_AT_VP:
            for col_id, x in enumerate(xs):
                for row_id, y in enumerate(ys):
                    empty_agg.values[row_id, col_id] = elev_at_vp
                    v = viewshed(raster=empty_agg, x=x, y=y,
                                 observer_elev=obs_elev)
                    # angle at viewpoint is always 180
                    assert v[row_id, col_id] == 180

                    if obs_elev + elev_at_vp >= 0 and \
                            obs_elev >= abs(elev_at_vp):
                        # all cells are visible
                        assert (v.values > -1).all()

                    b1, b2 = get_matrices(row_id, col_id, H, W)
                    m1 = v.values[b1[0]:b1[2], b1[1]:b1[3]]
                    m2 = v.values[b2[0]:b2[2], b2[1]:b2[3]]

                    assert check_symmetric(m1)
                    assert check_symmetric(m2[::-1])

                    # empty image for next uses
                    empty_agg.values[row_id, col_id] = 0



import numpy as np

import xarray as xr
from xarray import DataArray

from xrspatial.utils import ngjit

from typing import Optional

# TODO: change parameters to take agg instead of height / width


@ngjit
def _finish_bump(width, height, locs, heights, spread):
    out = np.zeros((height, width))
    rows, cols = out.shape
    s = spread ** 2  # removed sqrt for perf.
    for i in range(len(heights)):
        x = locs[i][0]
        y = locs[i][1]
        z = heights[i]
        out[y, x] = out[y, x] + z
        if s > 0:
            for nx in range(max(x - spread, 0), min(x + spread, width)):
                for ny in range(max(y - spread, 0), min(y + spread, height)):
                    d2 = (nx - x) * (nx - x) + (ny - y) * (ny - y)
                    if d2 <= s:
                        out[ny, nx] = out[ny, nx] + (out[y, x] * (d2 / s))
    return out


def bump(width: int,
         height: int,
         count: Optional[int] = None,
         height_func=None,
         spread: int = 1) -> xr.DataArray:
    """
    Generate a simple bump map to simulate the appearance of land
    features.

    Using a user-defined height function, determines at what elevation
    a specific bump height is acceptable. Bumps of number `count` are
    applied over the area `width` x `height`.

    Parameters
    ----------
    width : int
        Total width, in pixels, of the image.
    height : int
        Total height, in pixels, of the image.
    count : int
        Number of bumps to generate.
    height_func : function which takes x, y and returns a height value
        Function used to apply varying bump heights to different
        elevations.
    spread : int, default=1
        Number of pixels to spread on all sides.

    Returns
    -------
    bump_agg : xarray.DataArray
        2D aggregate array of calculated bump heights.

    References
    ----------
        - ICA: http://www.mountaincartography.org/mt_hood/pdfs/nighbert_bump1.pdf # noqa

    Examples
    --------
    .. plot::
       :include-source:

        import datashader as ds
        import matplotlib.pyplot as plt
        import numpy as np
        from xrspatial import generate_terrain, bump
        from functools import partial

        # Create Canvas
        W = 500
        H = 300
        cvs = ds.Canvas(plot_width = W,
                        plot_height = H,
                        x_range = (-20e6, 20e6),
                        y_range = (-20e6, 20e6))

        # Generate Example Terrain
        terrain_agg = generate_terrain(canvas = cvs)

        # Edit Attributes
        terrain_agg = terrain_agg.assign_attrs(
            {
                'Description': 'Example Terrain',
                'units': 'km',
                'Max Elevation': '4000',
            }
        )
        
        terrain_agg = terrain_agg.rename({'x': 'lon', 'y': 'lat'})
        terrain_agg = terrain_agg.rename('Elevation')

        # Create Height Function
        def heights(locations, src, src_range, height = 20):
            num_bumps = locations.shape[0]
            out = np.zeros(num_bumps, dtype = np.uint16)
            for r in range(0, num_bumps):
                loc = locations[r]
                x = loc[0]
                y = loc[1]
                val = src[y, x]
                if val >= src_range[0] and val < src_range[1]:
                    out[r] = height
            return out

        # Create Bump Map Aggregate Array
        bump_count = 10000
        src = terrain_agg.data

        # Short Bumps from z = 1000 to z = 1300
        bump_agg = bump(width = W, height = H, count = bump_count,
                        height_func = partial(heights, src = src,
                                            src_range = (1000, 1300),
                                            height = 5))

        # Tall Bumps from z = 1300 to z = 1700
        bump_agg += bump(width = W, height = H, count = bump_count // 2,
                        height_func = partial(heights, src = src,
                                            src_range = (1300, 1700),
                                            height=20))

        # Short Bumps from z = 1700 to z = 2000
        bump_agg += bump(width = W, height = H, count = bump_count // 3,
                        height_func = partial(heights, src = src,
                                            src_range = (1700, 2000),
                                            height=5))
        # Edit Attributes
        bump_agg = bump_agg.assign_attrs({'Description': 'Example Bump Map',
                                          'units': 'km'})

        bump_agg = bump_agg.rename('Bump Height')

        # Rename Coordinates
        bump_agg = bump_agg.assign_coords({'x': terrain_agg.coords['lon'].data,
                                           'y': terrain_agg.coords['lat'].data})

        # Remove zeros
        bump_agg.data[bump_agg.data == 0] = np.nan

        # Plot Terrain
        terrain_agg.plot(cmap = 'terrain', aspect = 2, size = 4)
        plt.title("Terrain")
        plt.ylabel("latitude")
        plt.xlabel("longitude")

        # Plot Bump Map
        bump_agg.plot(cmap = 'summer', aspect = 2, size = 4)
        plt.title("Bump Map")
        plt.ylabel("latitude")
        plt.xlabel("longitude")

    .. sourcecode:: python

        >>> print(terrain_agg[200:203, 200:202])
        <xarray.DataArray 'Elevation' (lat: 3, lon: 2)>
        array([[1264.02249454, 1261.94748873],
               [1285.37061171, 1282.48046696],
               [1306.02305679, 1303.40657515]])
        Coordinates:
          * lon      (lon) float64 -3.96e+06 -3.88e+06
          * lat      (lat) float64 6.733e+06 6.867e+06 7e+06
        Attributes:
            res:            1
            Description:    Example Terrain
            units:          km
            Max Elevation:  4000

    .. sourcecode:: python

        >>> print(bump_agg[200:205, 200:206])
        <xarray.DataArray 'Bump Height' (y: 5, x: 6)>
        array([[nan, nan, nan, nan,  5.,  5.],
               [nan, nan, nan, nan, nan,  5.],
               [nan, nan, nan, nan, nan, nan],
               [nan, nan, nan, nan, nan, nan],
               [nan, nan, nan, nan, nan, nan]])
        Coordinates:
          * x        (x) float64 -3.96e+06 -3.88e+06 -3.8e+06 ... -3.64e+06 -3.56e+06
          * y        (y) float64 6.733e+06 6.867e+06 7e+06 7.133e+06 7.267e+06
        Attributes:
            res:          1
            Description:  Example Bump Map
            units:        km
    """
    linx = range(width)
    liny = range(height)

    if count is None:
        count = width * height // 10

    if height_func is None:
        height_func = lambda bumps: np.ones(len(bumps)) # noqa

    # create 2d array of random x, y for bump locations
    locs = np.empty((count, 2), dtype=np.uint16)
    locs[:, 0] = np.random.choice(linx, count)
    locs[:, 1] = np.random.choice(liny, count)

    heights = height_func(locs)

    bumps = _finish_bump(width, height, locs, heights, spread)
    return DataArray(bumps, dims=['y', 'x'], attrs=dict(res=1))
